from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class TIME(DefinedNamespace):
    """
    OWL-Time

    Generated from: http://www.w3.org/2006/time#
    Date: 2020-05-26 14:20:10.531265

    """

    # http://www.w3.org/2000/01/rdf-schema#Datatype
    generalDay: URIRef  # Day of month - formulated as a text string with a pattern constraint to reproduce the same lexical form as gDay, except that values up to 99 are permitted, in order to support calendars with more than 31 days in a month.  Note that the value-space is not defined, so a generic OWL2 processor cannot compute ordering relationships of values of this type.
    generalMonth: URIRef  # Month of year - formulated as a text string with a pattern constraint to reproduce the same lexical form as gMonth, except that values up to 20 are permitted, in order to support calendars with more than 12 months in the year.  Note that the value-space is not defined, so a generic OWL2 processor cannot compute ordering relationships of values of this type.
    generalYear: URIRef  # Year number - formulated as a text string with a pattern constraint to reproduce the same lexical form as gYear, but not restricted to values from the Gregorian calendar.  Note that the value-space is not defined, so a generic OWL2 processor cannot compute ordering relationships of values of this type.

    # http://www.w3.org/2002/07/owl#Class
    DateTimeDescription: URIRef  # Description of date and time structured with separate values for the various elements of a calendar-clock system. The temporal reference system is fixed to Gregorian Calendar, and the range of year, month, day properties restricted to corresponding XML Schema types xsd:gYear, xsd:gMonth and xsd:gDay, respectively.
    DateTimeInterval: URIRef  # DateTimeInterval is a subclass of ProperInterval, defined using the multi-element DateTimeDescription.
    DayOfWeek: URIRef  # The day of week
    Duration: URIRef  # Duration of a temporal extent expressed as a number scaled by a temporal unit
    DurationDescription: URIRef  # Description of temporal extent structured with separate values for the various elements of a calendar-clock system. The temporal reference system is fixed to Gregorian Calendar, and the range of each of the numeric properties is restricted to xsd:decimal
    GeneralDateTimeDescription: URIRef  # Description of date and time structured with separate values for the various elements of a calendar-clock system
    GeneralDurationDescription: URIRef  # Description of temporal extent structured with separate values for the various elements of a calendar-clock system.
    Instant: URIRef  # A temporal entity with zero extent or duration
    Interval: URIRef  # A temporal entity with an extent or duration
    MonthOfYear: URIRef  # The month of the year
    ProperInterval: URIRef  # A temporal entity with non-zero extent or duration, i.e. for which the value of the beginning and end are different
    TRS: URIRef  # A temporal reference system, such as a temporal coordinate system (with an origin, direction, and scale), a calendar-clock combination, or a (possibly hierarchical) ordinal system.   This is a stub class, representing the set of all temporal reference systems.
    TemporalDuration: URIRef  # Time extent; duration of a time interval separate from its particular start position
    TemporalEntity: URIRef  # A temporal interval or instant.
    TemporalPosition: URIRef  # A position on a time-line
    TemporalUnit: URIRef  # A standard duration, which provides a scale factor for a time extent, or the granularity or precision for a time position.
    TimePosition: URIRef  # A temporal position described using either a (nominal) value from an ordinal reference system, or a (numeric) value in a temporal coordinate system.
    TimeZone: URIRef  # A Time Zone specifies the amount by which the local time is offset from UTC.  	A time zone is usually denoted geographically (e.g. Australian Eastern Daylight Time), with a constant value in a given region.  The region where it applies and the offset from UTC are specified by a locally recognised governing authority.

    # http://www.w3.org/2002/07/owl#DatatypeProperty
    day: URIRef  # Day position in a calendar-clock system.  The range of this property is not specified, so can be replaced by any specific representation of a calendar day from any calendar.
    dayOfYear: URIRef  # The number of the day within the year
    days: URIRef  # length of, or element of the length of, a temporal extent expressed in days
    hasXSDDuration: URIRef  # Extent of a temporal entity, expressed using xsd:duration
    hour: URIRef  # Hour position in a calendar-clock system.
    hours: URIRef  # length of, or element of the length of, a temporal extent expressed in hours
    inXSDDate: URIRef  # Position of an instant, expressed using xsd:date
    inXSDDateTimeStamp: (
        URIRef  # Position of an instant, expressed using xsd:dateTimeStamp
    )
    inXSDgYear: URIRef  # Position of an instant, expressed using xsd:gYear
    inXSDgYearMonth: URIRef  # Position of an instant, expressed using xsd:gYearMonth
    minute: URIRef  # Minute position in a calendar-clock system.
    minutes: URIRef  # length, or element of, a temporal extent expressed in minutes
    month: URIRef  # Month position in a calendar-clock system.  The range of this property is not specified, so can be replaced by any specific representation of a calendar month from any calendar.
    months: URIRef  # length of, or element of the length of, a temporal extent expressed in months
    nominalPosition: URIRef  # The (nominal) value indicating temporal position in an ordinal reference system
    numericDuration: URIRef  # Value of a temporal extent expressed as a decimal number scaled by a temporal unit
    numericPosition: URIRef  # The (numeric) value indicating position within a temporal coordinate system
    second: URIRef  # Second position in a calendar-clock system.
    seconds: URIRef  # length of, or element of the length of, a temporal extent expressed in seconds
    week: URIRef  # Week number within the year.
    weeks: URIRef  # length of, or element of the length of, a temporal extent expressed in weeks
    year: URIRef  # Year position in a calendar-clock system.  The range of this property is not specified, so can be replaced by any specific representation of a calendar year from any calendar.
    years: URIRef  # length of, or element of the length of, a temporal extent expressed in years

    # http://www.w3.org/2002/07/owl#DeprecatedClass
    January: URIRef  # January
    Year: URIRef  # Year duration

    # http://www.w3.org/2002/07/owl#DeprecatedProperty
    inXSDDateTime: URIRef  # Position of an instant, expressed using xsd:dateTime
    xsdDateTime: URIRef  # Value of DateTimeInterval expressed as a compact value.

    # http://www.w3.org/2002/07/owl#FunctionalProperty
    hasTRS: URIRef  # The temporal reference system used by a temporal position or extent description.

    # http://www.w3.org/2002/07/owl#ObjectProperty
    after: URIRef  # Gives directionality to time. If a temporal entity T1 is after another temporal entity T2, then the beginning of T1 is after the end of T2.
    dayOfWeek: (
        URIRef  # The day of week, whose value is a member of the class time:DayOfWeek
    )
    hasBeginning: URIRef  # Beginning of a temporal entity.
    hasDateTimeDescription: URIRef  # Value of DateTimeInterval expressed as a structured value. The beginning and end of the interval coincide with the limits of the shortest element in the description.
    hasDuration: URIRef  # Duration of a temporal entity, event or activity, or thing, expressed as a scaled value
    hasDurationDescription: URIRef  # Duration of a temporal entity, expressed using a structured description
    hasEnd: URIRef  # End of a temporal entity.
    hasTemporalDuration: URIRef  # Duration of a temporal entity.
    hasTime: URIRef  # Supports the association of a temporal entity (instant or interval) to any thing
    inDateTime: (
        URIRef  # Position of an instant, expressed using a structured description
    )
    inTemporalPosition: URIRef  # Position of a time instant
    inTimePosition: URIRef  # Position of a time instant expressed as a TimePosition
    inside: URIRef  # An instant that falls inside the interval. It is not intended to include beginnings and ends of intervals.
    intervalAfter: URIRef  # If a proper interval T1 is intervalAfter another proper interval T2, then the beginning of T1 is after the end of T2.
    intervalBefore: URIRef  # If a proper interval T1 is intervalBefore another proper interval T2, then the end of T1 is before the beginning of T2.
    intervalContains: URIRef  # If a proper interval T1 is intervalContains another proper interval T2, then the beginning of T1 is before the beginning of T2, and the end of T1 is after the end of T2.
    intervalDisjoint: URIRef  # If a proper interval T1 is intervalDisjoint another proper interval T2, then the beginning of T1 is after the end of T2, or the end of T1 is before the beginning of T2, i.e. the intervals do not overlap in any way, but their ordering relationship is not known.
    intervalDuring: URIRef  # If a proper interval T1 is intervalDuring another proper interval T2, then the beginning of T1 is after the beginning of T2, and the end of T1 is before the end of T2.
    intervalEquals: URIRef  # If a proper interval T1 is intervalEquals another proper interval T2, then the beginning of T1 is coincident with the beginning of T2, and the end of T1 is coincident with the end of T2.
    intervalFinishedBy: URIRef  # If a proper interval T1 is intervalFinishedBy another proper interval T2, then the beginning of T1 is before the beginning of T2, and the end of T1 is coincident with the end of T2.
    intervalFinishes: URIRef  # If a proper interval T1 is intervalFinishes another proper interval T2, then the beginning of T1 is after the beginning of T2, and the end of T1 is coincident with the end of T2.
    intervalIn: URIRef  # If a proper interval T1 is intervalIn another proper interval T2, then the beginning of T1 is after the beginning of T2 or is coincident with the beginning of T2, and the end of T1 is before the end of T2, or is coincident with the end of T2, except that end of T1 may not be coincident with the end of T2 if the beginning of T1 is coincident with the beginning of T2.
    intervalMeets: URIRef  # If a proper interval T1 is intervalMeets another proper interval T2, then the end of T1 is coincident with the beginning of T2.
    intervalMetBy: URIRef  # If a proper interval T1 is intervalMetBy another proper interval T2, then the beginning of T1 is coincident with the end of T2.
    intervalOverlappedBy: URIRef  # If a proper interval T1 is intervalOverlappedBy another proper interval T2, then the beginning of T1 is after the beginning of T2, the beginning of T1 is before the end of T2, and the end of T1 is after the end of T2.
    intervalOverlaps: URIRef  # If a proper interval T1 is intervalOverlaps another proper interval T2, then the beginning of T1 is before the beginning of T2, the end of T1 is after the beginning of T2, and the end of T1 is before the end of T2.
    intervalStartedBy: URIRef  # If a proper interval T1 is intervalStarted another proper interval T2, then the beginning of T1 is coincident with the beginning of T2, and the end of T1 is after the end of T2.
    intervalStarts: URIRef  # If a proper interval T1 is intervalStarts another proper interval T2, then the beginning of T1 is coincident with the beginning of T2, and the end of T1 is before the end of T2.
    monthOfYear: URIRef  # The month of the year, whose value is a member of the class time:MonthOfYear
    timeZone: URIRef  # The time zone for clock elements in the temporal position
    unitType: URIRef  # The temporal unit which provides the precision of a date-time value or scale of a temporal extent

    # http://www.w3.org/2002/07/owl#TransitiveProperty
    before: URIRef  # Gives directionality to time. If a temporal entity T1 is before another temporal entity T2, then the end of T1 is before the beginning of T2. Thus, "before" can be considered to be basic to instants and derived for intervals.

    # http://www.w3.org/2006/time#DayOfWeek
    Friday: URIRef  # Friday
    Monday: URIRef  # Monday
    Saturday: URIRef  # Saturday
    Sunday: URIRef  # Sunday
    Thursday: URIRef  # Thursday
    Tuesday: URIRef  # Tuesday
    Wednesday: URIRef  # Wednesday

    # http://www.w3.org/2006/time#TemporalUnit
    unitDay: URIRef  # day
    unitHour: URIRef  # hour
    unitMinute: URIRef  # minute
    unitMonth: URIRef  # month
    unitSecond: URIRef  # second
    unitWeek: URIRef  # week
    unitYear: URIRef  # year

    _NS = Namespace("http://www.w3.org/2006/time#")



"""Parsers and serializers for SPARQL Result formats

All builtin [SPARQL result serializers][rdflib.query.ResultSerializer] available
after [a query][rdflib.graph.Graph.query] are listed here.
The serializer can be chosen during [serialization of a result][rdflib.query.Result.serialize].

```
import rdflib

result : rdflib.query.Result
g = rdflib.Graph().parse(format="ttl", data=\"\"\"
    @base <http://example.org/> .
    <MyObject> a <MyClass>.
\"\"\")
result = g.query("SELECT ?x ?cls WHERE {?x a ?cls}")
print(result.serialize(format="txt").decode("utf8"))
```

Registration of builtin and external plugins
is described in [`rdflib.plugin`][rdflib.plugin].
"""



"""

This wrapper intercepts calls through the store interface and implements
thread-safe logging of destructive operations (adds / removes) in reverse.
This is persisted on the store instance and the reverse operations are
executed In order to return the store to the state it was when the transaction
began Since the reverse operations are persisted on the store, the store
itself acts as a transaction.

Calls to commit or rollback, flush the list of reverse operations This
provides thread-safe atomicity and isolation (assuming concurrent operations
occur with different store instances), but no durability (transactions are
persisted in memory and won't be available to reverse operations after the
system fails): A and I out of ACID.

"""

from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Any, Generator, Iterator, List, Optional, Tuple, Union

from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.store import Store

if TYPE_CHECKING:
    from rdflib.graph import (
        _ContextIdentifierType,
        _ContextType,
        _ObjectType,
        _PredicateType,
        _SubjectType,
        _TriplePatternType,
        _TripleType,
    )
    from rdflib.query import Result
    from rdflib.term import URIRef


destructiveOpLocks = {  # noqa: N816
    "add": None,
    "remove": None,
}


class AuditableStore(Store):
    """A store that logs destructive operations (add/remove) in reverse order."""

    def __init__(self, store: Store):
        self.store = store
        self.context_aware = store.context_aware
        # NOTE: this store can't be formula_aware as it doesn't have enough
        # info to reverse the removal of a quoted statement
        self.formula_aware = False  # store.formula_aware
        self.transaction_aware = True  # This is only half true
        self.reverseOps: List[
            Tuple[
                Optional[_SubjectType],
                Optional[_PredicateType],
                Optional[_ObjectType],
                Optional[_ContextIdentifierType],
                str,
            ]
        ] = []
        self.rollbackLock = threading.RLock()

    def open(
        self, configuration: Union[str, tuple[str, str]], create: bool = True
    ) -> Optional[int]:
        return self.store.open(configuration, create)

    def close(self, commit_pending_transaction: bool = False) -> None:
        self.store.close()

    def destroy(self, configuration: str) -> None:
        self.store.destroy(configuration)

    def query(self, *args: Any, **kw: Any) -> Result:
        return self.store.query(*args, **kw)

    def add(
        self, triple: _TripleType, context: _ContextType, quoted: bool = False
    ) -> None:
        (s, p, o) = triple
        lock = destructiveOpLocks["add"]
        lock = lock if lock else threading.RLock()
        with lock:
            context = (
                context.__class__(self.store, context.identifier)
                if context is not None
                else None
            )
            ctxId = context.identifier if context is not None else None  # noqa: N806
            if list(self.store.triples(triple, context)):
                return  # triple already in store, do nothing
            self.reverseOps.append((s, p, o, ctxId, "remove"))
            try:
                self.reverseOps.remove((s, p, o, ctxId, "add"))
            except ValueError:
                pass
            self.store.add((s, p, o), context, quoted)

    def remove(
        self, spo: _TriplePatternType, context: Optional[_ContextType] = None
    ) -> None:
        subject, predicate, object_ = spo
        lock = destructiveOpLocks["remove"]
        lock = lock if lock else threading.RLock()
        with lock:
            # Need to determine which quads will be removed if any term is a
            # wildcard
            context = (
                context.__class__(self.store, context.identifier)
                if context is not None
                else None
            )
            ctxId = context.identifier if context is not None else None  # noqa: N806
            if None in [subject, predicate, object_, context]:
                if ctxId:
                    # type error: Item "None" of "Optional[Graph]" has no attribute "triples"
                    for s, p, o in context.triples((subject, predicate, object_)):  # type: ignore[union-attr]
                        try:
                            self.reverseOps.remove((s, p, o, ctxId, "remove"))
                        except ValueError:
                            self.reverseOps.append((s, p, o, ctxId, "add"))
                else:
                    for s, p, o, ctx in ConjunctiveGraph(self.store).quads(
                        (subject, predicate, object_)
                    ):
                        try:
                            # type error: Item "None" of "Optional[Graph]" has no attribute "identifier"
                            self.reverseOps.remove((s, p, o, ctx.identifier, "remove"))  # type: ignore[union-attr]
                        except ValueError:
                            # type error: Item "None" of "Optional[Graph]" has no attribute "identifier"
                            self.reverseOps.append((s, p, o, ctx.identifier, "add"))  # type: ignore[union-attr]
            else:
                if not list(self.triples((subject, predicate, object_), context)):
                    return  # triple not present in store, do nothing
                try:
                    self.reverseOps.remove(
                        (subject, predicate, object_, ctxId, "remove")
                    )
                except ValueError:
                    self.reverseOps.append((subject, predicate, object_, ctxId, "add"))
            self.store.remove((subject, predicate, object_), context)

    def triples(
        self, triple: _TriplePatternType, context: Optional[_ContextType] = None
    ) -> Iterator[Tuple[_TripleType, Iterator[Optional[_ContextType]]]]:
        (su, pr, ob) = triple
        context = (
            context.__class__(self.store, context.identifier)
            if context is not None
            else None
        )
        for (s, p, o), cg in self.store.triples((su, pr, ob), context):
            yield (s, p, o), cg

    def __len__(self, context: Optional[_ContextType] = None):
        context = (
            context.__class__(self.store, context.identifier)
            if context is not None
            else None
        )
        return self.store.__len__(context)

    def contexts(
        self, triple: Optional[_TripleType] = None
    ) -> Generator[_ContextType, None, None]:
        for ctx in self.store.contexts(triple):
            yield ctx

    def bind(self, prefix: str, namespace: URIRef, override: bool = True) -> None:
        self.store.bind(prefix, namespace, override=override)

    def prefix(self, namespace: URIRef) -> Optional[str]:
        return self.store.prefix(namespace)

    def namespace(self, prefix: str) -> Optional[URIRef]:
        return self.store.namespace(prefix)

    def namespaces(self) -> Iterator[Tuple[str, URIRef]]:
        return self.store.namespaces()

    def commit(self) -> None:
        self.reverseOps = []

    def rollback(self) -> None:
        # Acquire Rollback lock and apply reverse operations in the forward
        # order
        with self.rollbackLock:
            for subject, predicate, obj, context, op in self.reverseOps:
                if op == "add":
                    # type error: Argument 2 to "Graph" has incompatible type "Optional[Node]"; expected "Union[IdentifiedNode, str, None]"
                    self.store.add(
                        (subject, predicate, obj), Graph(self.store, context)  # type: ignore[arg-type]
                    )
                else:
                    self.store.remove(
                        (subject, predicate, obj), Graph(self.store, context)
                    )

            self.reverseOps = []



"""Module for builtin RDF parser

All builtin [RDF parser][rdflib.parser.Parser] available
in [RDFLib Graphs method `parse`][rdflib.graph.Graph.parse] are listed here.

Registration of builtin and external plugins
is described in [`rdflib.plugin`][rdflib.plugin].
"""



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class DC(DefinedNamespace):
    """
    Dublin Core Metadata Element Set, Version 1.1

    Generated from: https://www.dublincore.org/specifications/dublin-core/dcmi-terms/dublin_core_elements.ttl
    Date: 2020-05-26 14:19:58.671906

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    contributor: (
        URIRef  # An entity responsible for making contributions to the resource.
    )
    coverage: URIRef  # The spatial or temporal topic of the resource, spatial applicability of the resource, or jurisdiction under which the resource is relevant.
    creator: URIRef  # An entity primarily responsible for making the resource.
    date: URIRef  # A point or period of time associated with an event in the lifecycle of the resource.
    description: URIRef  # An account of the resource.
    format: URIRef  # The file format, physical medium, or dimensions of the resource.
    identifier: (
        URIRef  # An unambiguous reference to the resource within a given context.
    )
    language: URIRef  # A language of the resource.
    publisher: URIRef  # An entity responsible for making the resource available.
    relation: URIRef  # A related resource.
    rights: URIRef  # Information about rights held in and over the resource.
    source: URIRef  # A related resource from which the described resource is derived.
    subject: URIRef  # The topic of the resource.
    title: URIRef  # A name given to the resource.
    type: URIRef  # The nature or genre of the resource.

    _NS = Namespace("http://purl.org/dc/elements/1.1/")



r"""
This module implements the SPARQL 1.1 Property path operators, as
defined in:
[http://www.w3.org/TR/sparql11-query/#propertypaths](http://www.w3.org/TR/sparql11-query/#propertypaths)

In SPARQL the syntax is as follows:

| Syntax              | Matches                                                                 |
|---------------------|-------------------------------------------------------------------------|
| `iri`               | An IRI. A path of length one.                                           |
| `^elt`              | Inverse path (object to subject).                                       |
| `elt1 / elt2`       | A sequence path of `elt1` followed by `elt2`.                           |
| `elt1 \| elt2`      | An alternative path of `elt1` or `elt2` (all possibilities are tried).  |
| `elt*`              | A path that connects subject and object by zero or more matches of `elt`.|
| `elt+`              | A path that connects subject and object by one or more matches of `elt`.|
| `elt?`              | A path that connects subject and object by zero or one matches of `elt`.|
| `!iri` or <br> `!(iri1 \| ... \| irin)` | Negated property set. An IRI not among `iri1` to `irin`. <br> `!iri` is short for `!(iri)`. |
| `!^iri` or <br> `!(^iri1 \| ... \| ^irin)` | Negated reverse property set. Excludes `^iri1` to `^irin` as reverse paths. <br> `!^iri` is short for `!(^iri)`. |
| `!(iri1 \| ... \| irij \| ^irij+1 \| ... \| ^irin)` | A combination of forward and reverse properties in a negated property set. |
| `(elt)`             | A grouped path `elt`, where parentheses control precedence.             |

This module is used internally by the SPARQL engine, but the property paths
can also be used to query RDFLib Graphs directly.

Where possible the SPARQL syntax is mapped to Python operators, and property
path objects can be constructed from existing URIRefs.

```python
>>> from rdflib import Graph, Namespace
>>> from rdflib.namespace import FOAF

>>> ~FOAF.knows
Path(~http://xmlns.com/foaf/0.1/knows)

>>> FOAF.knows/FOAF.name
Path(http://xmlns.com/foaf/0.1/knows / http://xmlns.com/foaf/0.1/name)

>>> FOAF.name|FOAF.givenName
Path(http://xmlns.com/foaf/0.1/name | http://xmlns.com/foaf/0.1/givenName)

```

Modifiers (?, \*, +) are done using \* (the multiplication operator) and
the strings '\*', '?', '+', also defined as constants in this file.

```python
>>> FOAF.knows*OneOrMore
Path(http://xmlns.com/foaf/0.1/knows+)

```

The path objects can also be used with the normal graph methods.

First some example data:

```python
>>> g=Graph()

>>> g=g.parse(data='''
... @prefix : <ex:> .
...
... :a :p1 :c ; :p2 :f .
... :c :p2 :e ; :p3 :g .
... :g :p3 :h ; :p2 :j .
... :h :p3 :a ; :p2 :g .
...
... :q :px :q .
...
... ''', format='n3') # doctest: +ELLIPSIS

>>> e = Namespace('ex:')

```

Graph contains:

```python
>>> (e.a, e.p1/e.p2, e.e) in g
True

```

Graph generator functions, triples, subjects, objects, etc. :

```python
>>> list(g.objects(e.c, (e.p3*OneOrMore)/e.p2)) # doctest: +NORMALIZE_WHITESPACE
[rdflib.term.URIRef('ex:j'), rdflib.term.URIRef('ex:g'),
    rdflib.term.URIRef('ex:f')]

```

A more complete set of tests:

```python
>>> list(eval_path(g, (None, e.p1/e.p2, None)))==[(e.a, e.e)]
True
>>> list(eval_path(g, (e.a, e.p1|e.p2, None)))==[(e.a,e.c), (e.a,e.f)]
True
>>> list(eval_path(g, (e.c, ~e.p1, None))) == [ (e.c, e.a) ]
True
>>> list(eval_path(g, (e.a, e.p1*ZeroOrOne, None))) == [(e.a, e.a), (e.a, e.c)]
True
>>> list(eval_path(g, (e.c, e.p3*OneOrMore, None))) == [
...     (e.c, e.g), (e.c, e.h), (e.c, e.a)]
True
>>> list(eval_path(g, (e.c, e.p3*ZeroOrMore, None))) == [(e.c, e.c),
...     (e.c, e.g), (e.c, e.h), (e.c, e.a)]
True
>>> list(eval_path(g, (e.a, -e.p1, None))) == [(e.a, e.f)]
True
>>> list(eval_path(g, (e.a, -(e.p1|e.p2), None))) == []
True
>>> list(eval_path(g, (e.g, -~e.p2, None))) == [(e.g, e.j)]
True
>>> list(eval_path(g, (e.e, ~(e.p1/e.p2), None))) == [(e.e, e.a)]
True
>>> list(eval_path(g, (e.a, e.p1/e.p3/e.p3, None))) == [(e.a, e.h)]
True

>>> list(eval_path(g, (e.q, e.px*OneOrMore, None)))
[(rdflib.term.URIRef('ex:q'), rdflib.term.URIRef('ex:q'))]

>>> list(eval_path(g, (None, e.p1|e.p2, e.c)))
[(rdflib.term.URIRef('ex:a'), rdflib.term.URIRef('ex:c'))]

>>> list(eval_path(g, (None, ~e.p1, e.a))) == [ (e.c, e.a) ]
True
>>> list(eval_path(g, (None, e.p1*ZeroOrOne, e.c))) # doctest: +NORMALIZE_WHITESPACE
[(rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:c')),
 (rdflib.term.URIRef('ex:a'), rdflib.term.URIRef('ex:c'))]

>>> list(eval_path(g, (None, e.p3*OneOrMore, e.a))) # doctest: +NORMALIZE_WHITESPACE
[(rdflib.term.URIRef('ex:h'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:g'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:a'))]

>>> list(eval_path(g, (None, e.p3*ZeroOrMore, e.a))) # doctest: +NORMALIZE_WHITESPACE
[(rdflib.term.URIRef('ex:a'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:h'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:g'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:a'))]

>>> list(eval_path(g, (None, -e.p1, e.f))) == [(e.a, e.f)]
True
>>> list(eval_path(g, (None, -(e.p1|e.p2), e.c))) == []
True
>>> list(eval_path(g, (None, -~e.p2, e.j))) == [(e.g, e.j)]
True
>>> list(eval_path(g, (None, ~(e.p1/e.p2), e.a))) == [(e.e, e.a)]
True
>>> list(eval_path(g, (None, e.p1/e.p3/e.p3, e.h))) == [(e.a, e.h)]
True

>>> list(eval_path(g, (e.q, e.px*OneOrMore, None)))
[(rdflib.term.URIRef('ex:q'), rdflib.term.URIRef('ex:q'))]

>>> list(eval_path(g, (e.c, (e.p2|e.p3)*ZeroOrMore, e.j)))
[(rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:j'))]

```

No vars specified:

```python
>>> sorted(list(eval_path(g, (None, e.p3*OneOrMore, None)))) #doctest: +NORMALIZE_WHITESPACE
[(rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:g')),
 (rdflib.term.URIRef('ex:c'), rdflib.term.URIRef('ex:h')),
 (rdflib.term.URIRef('ex:g'), rdflib.term.URIRef('ex:a')),
 (rdflib.term.URIRef('ex:g'), rdflib.term.URIRef('ex:h')),
 (rdflib.term.URIRef('ex:h'), rdflib.term.URIRef('ex:a'))]

```
"""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from functools import total_ordering
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from rdflib.term import Node, URIRef

if TYPE_CHECKING:
    from rdflib._type_checking import _MulPathMod
    from rdflib.graph import Graph, _ObjectType, _PredicateType, _SubjectType
    from rdflib.namespace import NamespaceManager


# property paths

ZeroOrMore = "*"
OneOrMore = "+"
ZeroOrOne = "?"


def _n3(
    arg: Union[URIRef, Path], namespace_manager: Optional[NamespaceManager] = None
) -> str:
    if isinstance(arg, (SequencePath, AlternativePath)) and len(arg.args) > 1:
        return "(%s)" % arg.n3(namespace_manager)
    return arg.n3(namespace_manager)


@total_ordering
class Path(ABC):
    """Base class for all property paths."""

    __or__: Callable[[Path, Union[URIRef, Path]], AlternativePath]
    __invert__: Callable[[Path], InvPath]
    __neg__: Callable[[Path], NegatedPath]
    __truediv__: Callable[[Path, Union[URIRef, Path]], SequencePath]
    __mul__: Callable[[Path, str], MulPath]

    @abstractmethod
    def eval(
        self,
        graph: Graph,
        subj: Optional[_SubjectType] = None,
        obj: Optional[_ObjectType] = None,
    ) -> Iterator[Tuple[_SubjectType, _ObjectType]]: ...

    @abstractmethod
    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str: ...

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (Path, Node)):
            raise TypeError(
                "unorderable types: %s() < %s()" % (repr(self), repr(other))
            )
        return repr(self) < repr(other)


class InvPath(Path):
    def __init__(self, arg: Union[Path, URIRef]):
        self.arg = arg

    def eval(
        self,
        graph: Graph,
        subj: Optional[_SubjectType] = None,
        obj: Optional[_ObjectType] = None,
    ) -> Generator[Tuple[_ObjectType, _SubjectType], None, None]:
        for s, o in eval_path(graph, (obj, self.arg, subj)):
            yield o, s

    def __repr__(self) -> str:
        return "Path(~%s)" % (self.arg,)

    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
        return "^%s" % _n3(self.arg, namespace_manager)


class SequencePath(Path):
    def __init__(self, *args: Union[Path, URIRef]):
        self.args: List[Union[Path, URIRef]] = []
        for a in args:
            if isinstance(a, SequencePath):
                self.args += a.args
            else:
                self.args.append(a)

    def eval(
        self,
        graph: Graph,
        subj: Optional[_SubjectType] = None,
        obj: Optional[_ObjectType] = None,
    ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
        def _eval_seq(
            paths: List[Union[Path, URIRef]],
            subj: Optional[_SubjectType],
            obj: Optional[_ObjectType],
        ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
            if paths[1:]:
                for s, o in eval_path(graph, (subj, paths[0], None)):
                    for r in _eval_seq(paths[1:], o, obj):
                        yield s, r[1]

            else:
                for s, o in eval_path(graph, (subj, paths[0], obj)):
                    yield s, o

        def _eval_seq_bw(
            paths: List[Union[Path, URIRef]],
            subj: Optional[_SubjectType],
            obj: _ObjectType,
        ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
            if paths[:-1]:
                for s, o in eval_path(graph, (None, paths[-1], obj)):
                    for r in _eval_seq(paths[:-1], subj, s):
                        yield r[0], o

            else:
                for s, o in eval_path(graph, (subj, paths[0], obj)):
                    yield s, o

        if subj:
            return _eval_seq(self.args, subj, obj)
        elif obj:
            return _eval_seq_bw(self.args, subj, obj)
        else:  # no vars bound, we can start anywhere
            return _eval_seq(self.args, subj, obj)

    def __repr__(self) -> str:
        return "Path(%s)" % " / ".join(str(x) for x in self.args)

    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
        return "/".join(_n3(a, namespace_manager) for a in self.args)


class AlternativePath(Path):
    def __init__(self, *args: Union[Path, URIRef]):
        self.args: List[Union[Path, URIRef]] = []
        for a in args:
            if isinstance(a, AlternativePath):
                self.args += a.args
            else:
                self.args.append(a)

    def eval(
        self,
        graph: Graph,
        subj: Optional[_SubjectType] = None,
        obj: Optional[_ObjectType] = None,
    ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
        for x in self.args:
            for y in eval_path(graph, (subj, x, obj)):
                yield y

    def __repr__(self) -> str:
        return "Path(%s)" % " | ".join(str(x) for x in self.args)

    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
        return "|".join(_n3(a, namespace_manager) for a in self.args)


class MulPath(Path):
    def __init__(self, path: Union[Path, URIRef], mod: _MulPathMod):
        self.path = path
        self.mod = mod

        if mod == ZeroOrOne:
            self.zero = True
            self.more = False
        elif mod == ZeroOrMore:
            self.zero = True
            self.more = True
        elif mod == OneOrMore:
            self.zero = False
            self.more = True
        else:
            raise Exception("Unknown modifier %s" % mod)

    def eval(
        self,
        graph: Graph,
        subj: Optional[_SubjectType] = None,
        obj: Optional[_ObjectType] = None,
        first: bool = True,
    ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
        if self.zero and first:
            if subj and obj:
                if subj == obj:
                    yield subj, obj
            elif subj:
                yield subj, subj
            elif obj:
                yield obj, obj

        def _fwd(
            subj: Optional[_SubjectType] = None,
            obj: Optional[_ObjectType] = None,
            seen: Optional[Set[_SubjectType]] = None,
        ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
            # type error: Item "None" of "Optional[Set[Node]]" has no attribute "add"
            # type error: Argument 1 to "add" of "set" has incompatible type "Optional[Node]"; expected "Node"
            seen.add(subj)  # type: ignore[union-attr, arg-type]

            for s, o in eval_path(graph, (subj, self.path, None)):
                if not obj or o == obj:
                    yield s, o
                if self.more:
                    # type error: Unsupported right operand type for in ("Optional[Set[Node]]")
                    if o in seen:  # type: ignore[operator]
                        continue
                    for s2, o2 in _fwd(o, obj, seen):
                        yield s, o2

        def _bwd(
            subj: Optional[_SubjectType] = None,
            obj: Optional[_ObjectType] = None,
            seen: Optional[Set[_ObjectType]] = None,
        ) -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
            # type error: Item "None" of "Optional[Set[Node]]" has no attribute "add"
            # type error: Argument 1 to "add" of "set" has incompatible type "Optional[Node]"; expected "Node"
            seen.add(obj)  # type: ignore[union-attr, arg-type]

            for s, o in eval_path(graph, (None, self.path, obj)):
                if not subj or subj == s:
                    yield s, o
                if self.more:
                    # type error: Unsupported right operand type for in ("Optional[Set[Node]]")
                    if s in seen:  # type: ignore[operator]
                        continue

                    for s2, o2 in _bwd(None, s, seen):
                        yield s2, o

        def _all_fwd_paths() -> Generator[Tuple[_SubjectType, _ObjectType], None, None]:
            if self.zero:
                seen1 = set()
                # According to the spec, ALL nodes are possible solutions
                # (even literals)
                # we cannot do this without going through ALL triples
                # unless we keep an index of all terms somehow
                # but let's just hope this query doesn't happen very often...
                for s, o in graph.subject_objects(None):
                    if s not in seen1:
                        seen1.add(s)
                        yield s, s
                    if o not in seen1:
                        seen1.add(o)
                        yield o, o

            seen = set()
            for s, o in eval_path(graph, (None, self.path, None)):
                if not self.more:
                    yield s, o
                else:
                    if s not in seen:
                        seen.add(s)
                        f = list(_fwd(s, None, set()))
                        for s1, o1 in f:
                            assert s1 == s
                            yield s1, o1

        done = set()  # the spec does, by defn, not allow duplicates
        if subj:
            for x in _fwd(subj, obj, set()):
                if x not in done:
                    done.add(x)
                    yield x
        elif obj:
            for x in _bwd(subj, obj, set()):
                if x not in done:
                    done.add(x)
                    yield x
        else:
            for x in _all_fwd_paths():
                if x not in done:
                    done.add(x)
                    yield x

    def __repr__(self) -> str:
        return "Path(%s%s)" % (self.path, self.mod)

    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
        return "%s%s" % (_n3(self.path, namespace_manager), self.mod)


class NegatedPath(Path):
    def __init__(self, arg: Union[AlternativePath, InvPath, URIRef]):
        self.args: List[Union[URIRef, Path]]
        if isinstance(arg, (URIRef, InvPath)):
            self.args = [arg]
        elif isinstance(arg, AlternativePath):
            self.args = arg.args
        else:
            raise Exception(
                "Can only negate URIRefs, InvPaths or "
                + "AlternativePaths, not: %s" % (arg,)
            )

    def eval(self, graph, subj=None, obj=None):
        for s, p, o in graph.triples((subj, None, obj)):
            for a in self.args:
                if isinstance(a, URIRef):
                    if p == a:
                        break
                elif isinstance(a, InvPath):
                    if (o, a.arg, s) in graph:
                        break
                else:
                    raise Exception("Invalid path in NegatedPath: %s" % a)
            else:
                yield s, o

    def __repr__(self) -> str:
        return "Path(! %s)" % ",".join(str(x) for x in self.args)

    def n3(self, namespace_manager: Optional[NamespaceManager] = None) -> str:
        return "!(%s)" % ("|".join(_n3(arg, namespace_manager) for arg in self.args))


class PathList(list):
    pass


def path_alternative(self: Union[URIRef, Path], other: Union[URIRef, Path]):
    """
    alternative path
    """
    if not isinstance(other, (URIRef, Path)):
        raise Exception("Only URIRefs or Paths can be in paths!")
    return AlternativePath(self, other)


def path_sequence(self: Union[URIRef, Path], other: Union[URIRef, Path]):
    """
    sequence path
    """
    if not isinstance(other, (URIRef, Path)):
        raise Exception("Only URIRefs or Paths can be in paths!")
    return SequencePath(self, other)


def evalPath(  # noqa: N802
    graph: Graph,
    t: Tuple[
        Optional[_SubjectType],
        Union[None, Path, _PredicateType],
        Optional[_ObjectType],
    ],
) -> Iterator[Tuple[_SubjectType, _ObjectType]]:
    warnings.warn(
        DeprecationWarning(
            "rdflib.path.evalPath() is deprecated, use the (snake-cased) eval_path(). "
            "The mixed-case evalPath() function name is incompatible with PEP8 "
            "recommendations and will be replaced by eval_path() in rdflib 7.0.0."
        )
    )
    return eval_path(graph, t)


def eval_path(
    graph: Graph,
    t: Tuple[
        Optional[_SubjectType],
        Union[None, Path, _PredicateType],
        Optional[_ObjectType],
    ],
) -> Iterator[Tuple[_SubjectType, _ObjectType]]:
    return ((s, o) for s, p, o in graph.triples(t))


def mul_path(p: Union[URIRef, Path], mul: _MulPathMod) -> MulPath:
    """
    cardinality path
    """
    return MulPath(p, mul)


def inv_path(p: Union[URIRef, Path]) -> InvPath:
    """
    inverse path
    """
    return InvPath(p)


def neg_path(p: Union[URIRef, AlternativePath, InvPath]) -> NegatedPath:
    """
    negated path
    """
    return NegatedPath(p)


if __name__ == "__main__":
    pass
else:
    # monkey patch
    # (these cannot be directly in terms.py
    #  as it would introduce circular imports)

    URIRef.__or__ = path_alternative
    # ignore typing here as URIRef inherits from str,
    # which has an incompatible definition of __mul__.
    URIRef.__mul__ = mul_path  # type: ignore
    URIRef.__invert__ = inv_path
    URIRef.__neg__ = neg_path
    URIRef.__truediv__ = path_sequence

    Path.__invert__ = inv_path
    # type error: Incompatible types in assignment (expression has type "Callable[[Union[URIRef, AlternativePath, InvPath]], NegatedPath]", variable has type "Callable[[Path], NegatedPath]")
    Path.__neg__ = neg_path  # type: ignore[assignment]
    # type error: Incompatible types in assignment (expression has type "Callable[[Union[URIRef, Path], Literal['*', '+', '?']], MulPath]", variable has type "Callable[[Path, str], MulPath]")
    Path.__mul__ = mul_path  # type: ignore[assignment]
    Path.__or__ = path_alternative
    Path.__truediv__ = path_sequence



from __future__ import annotations

import base64
import copy
import logging
from io import BytesIO
from typing import TYPE_CHECKING, Optional, Tuple
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from rdflib.plugin import plugins
from rdflib.query import Result, ResultParser
from rdflib.term import BNode
from rdflib.util import FORMAT_MIMETYPE_MAP, RESPONSE_TABLE_FORMAT_MIMETYPE_MAP

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    import typing_extensions as te


class SPARQLConnectorException(Exception):  # noqa: N818
    pass


class SPARQLConnector:
    """
    this class deals with nitty gritty details of talking to a SPARQL server
    """

    def __init__(
        self,
        query_endpoint: Optional[str] = None,
        update_endpoint: Optional[str] = None,
        returnFormat: Optional[str] = "xml",  # noqa: N803
        method: te.Literal["GET", "POST", "POST_FORM"] = "GET",
        auth: Optional[Tuple[str, str]] = None,
        **kwargs,
    ):
        """
        auth, if present, must be a tuple of (username, password) used for Basic Authentication

        Any additional keyword arguments will be passed to to the request, and can be used to setup timeouts etc.
        """
        self._method: str
        self.returnFormat = returnFormat
        self.query_endpoint = query_endpoint
        self.update_endpoint = update_endpoint
        self.kwargs = kwargs
        self.method = method
        if auth is not None:
            if type(auth) is not tuple:
                raise SPARQLConnectorException("auth must be a tuple")
            if len(auth) != 2:
                raise SPARQLConnectorException("auth must be a tuple (user, password)")
            base64string = base64.b64encode(bytes("%s:%s" % auth, "ascii"))
            self.kwargs.setdefault("headers", {})
            self.kwargs["headers"].update(
                {"Authorization": "Basic %s" % base64string.decode("utf-8")}
            )

    @property
    def method(self) -> str:
        return self._method

    @method.setter
    def method(self, method: str) -> None:
        if method not in ("GET", "POST", "POST_FORM"):
            raise SPARQLConnectorException(
                'Method must be "GET", "POST", or "POST_FORM"'
            )

        self._method = method

    def query(
        self,
        query: str,
        default_graph: Optional[str] = None,
        named_graph: Optional[str] = None,
    ) -> Result:
        if not self.query_endpoint:
            raise SPARQLConnectorException("Query endpoint not set!")

        params = {}
        # this test ensures we don't have a useless (BNode) default graph URI, which calls to Graph().query() will add
        if default_graph is not None and type(default_graph) is not BNode:
            params["default-graph-uri"] = default_graph

        headers = {"Accept": self.response_mime_types()}

        args = copy.deepcopy(self.kwargs)

        # merge params/headers dicts
        args.setdefault("params", {})

        args.setdefault("headers", {})
        args["headers"].update(headers)

        if self.method == "GET":
            params["query"] = query
            args["params"].update(params)
            qsa = "?" + urlencode(args["params"])
            try:
                res = urlopen(
                    Request(self.query_endpoint + qsa, headers=args["headers"])
                )
            except Exception as e:  # noqa: F841
                raise ValueError(
                    "You did something wrong formulating either the URI or your SPARQL query"
                )
        elif self.method == "POST":
            args["headers"].update({"Content-Type": "application/sparql-query"})
            args["params"].update(params)
            qsa = "?" + urlencode(args["params"])
            try:
                res = urlopen(
                    Request(
                        self.query_endpoint + qsa,
                        data=query.encode(),
                        headers=args["headers"],
                    )
                )
            except HTTPError as e:
                # type error: Incompatible return value type (got "Tuple[int, str, None]", expected "Result")
                return e.code, str(e), None  # type: ignore[return-value]
        elif self.method == "POST_FORM":
            params["query"] = query
            args["params"].update(params)
            try:
                res = urlopen(
                    Request(
                        self.query_endpoint,
                        data=urlencode(args["params"]).encode(),
                        headers=args["headers"],
                    )
                )
            except HTTPError as e:
                # type error: Incompatible return value type (got "Tuple[int, str, None]", expected "Result")
                return e.code, str(e), None  # type: ignore[return-value]
        else:
            raise SPARQLConnectorException("Unknown method %s" % self.method)
        return Result.parse(
            BytesIO(res.read()), content_type=res.headers["Content-Type"].split(";")[0]
        )

    def update(
        self,
        query: str,
        default_graph: Optional[str] = None,
        named_graph: Optional[str] = None,
    ) -> None:
        if not self.update_endpoint:
            raise SPARQLConnectorException("Query endpoint not set!")

        params = {}

        if default_graph is not None:
            params["using-graph-uri"] = default_graph

        if named_graph is not None:
            params["using-named-graph-uri"] = named_graph

        headers = {
            "Accept": self.response_mime_types(),
            "Content-Type": "application/sparql-update; charset=UTF-8",
        }

        args = copy.deepcopy(self.kwargs)  # other QSAs

        args.setdefault("params", {})
        args["params"].update(params)
        args.setdefault("headers", {})
        args["headers"].update(headers)

        qsa = "?" + urlencode(args["params"])
        res = urlopen(  # noqa: F841
            Request(
                self.update_endpoint + qsa, data=query.encode(), headers=args["headers"]
            )
        )

    def response_mime_types(self) -> str:
        """Construct a HTTP-Header Accept field to reflect the supported mime types.

        If the return_format parameter is set, the mime types are restricted to these accordingly.
        """
        sparql_format_mimetype_map = {
            k: FORMAT_MIMETYPE_MAP.get(k, [])
            + RESPONSE_TABLE_FORMAT_MIMETYPE_MAP.get(k, [])
            for k in list(FORMAT_MIMETYPE_MAP.keys())
            + list(RESPONSE_TABLE_FORMAT_MIMETYPE_MAP.keys())
        }

        supported_formats = set()
        for plugin in plugins(name=self.returnFormat, kind=ResultParser):
            if "/" not in plugin.name:
                supported_formats.update(
                    sparql_format_mimetype_map.get(plugin.name, [])
                )
            else:
                supported_formats.add(plugin.name)
        return ", ".join(supported_formats)


__all__ = ["SPARQLConnector", "SPARQLConnectorException"]



"""
This file provides a single function `serialize_in_chunks()` which can serialize a
Graph into a number of NT files with a maximum number of triples or maximum file size.

There is an option to preserve any prefixes declared for the original graph in the first
file, which will be a Turtle file.
"""

from __future__ import annotations

from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO, Generator, Optional, Tuple

from rdflib.graph import Graph
from rdflib.plugins.serializers.nt import _nt_row

# from rdflib.term import Literal

# if TYPE_CHECKING:
#     from rdflib.graph import _TriplePatternType

__all__ = ["serialize_in_chunks"]


def serialize_in_chunks(
    g: Graph,
    max_triples: int = 10000,
    max_file_size_kb: Optional[int] = None,
    file_name_stem: str = "chunk",
    output_dir: Optional[Path] = None,
    write_prefixes: bool = False,
) -> None:
    """Serializes a given Graph into a series of n-triples with a given length.

    Args:
        g: The graph to serialize.
        max_file_size_kb: Maximum size per NT file in kB (1,000 bytes)
            Equivalent to ~6,000 triples, depending on Literal sizes.
        max_triples: Maximum size per NT file in triples
            Equivalent to lines in file.
            If both this parameter and max_file_size_kb are set, max_file_size_kb will be used.
        file_name_stem: Prefix of each file name.
            e.g. "chunk" = chunk_000001.nt, chunk_000002.nt...
        output_dir: The directory you want the files to be written to.
        write_prefixes: The first file created is a Turtle file containing original graph prefixes.

    See `../test/test_tools/test_chunk_serializer.py` for examples of this in use.
    """

    if output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.is_dir():
        raise ValueError(
            "If you specify an output_dir, it must actually be a directory!"
        )

    @contextmanager
    def _start_new_file(file_no: int) -> Generator[Tuple[Path, BinaryIO], None, None]:
        if TYPE_CHECKING:
            # this is here because mypy gets a bit confused
            assert output_dir is not None
        fp = Path(output_dir) / f"{file_name_stem}_{str(file_no).zfill(6)}.nt"
        with open(fp, "wb") as fh:
            yield fp, fh

    def _serialize_prefixes(g: Graph) -> str:
        pres = []
        for k, v in g.namespace_manager.namespaces():
            pres.append(f"PREFIX {k}: <{v}>")

        return "\n".join(sorted(pres)) + "\n"

    if write_prefixes:
        with open(
            Path(output_dir) / f"{file_name_stem}_000000.ttl", "w", encoding="utf-8"
        ) as fh:
            fh.write(_serialize_prefixes(g))

    bytes_written = 0
    with ExitStack() as xstack:
        if max_file_size_kb is not None:
            max_file_size = max_file_size_kb * 1000
            file_no = 1 if write_prefixes else 0
            for i, t in enumerate(g.triples((None, None, None))):
                row_bytes = _nt_row(t).encode("utf-8")
                if len(row_bytes) > max_file_size:
                    raise ValueError(
                        # type error: Unsupported operand types for / ("bytes" and "int")
                        f"cannot write triple {t!r} as it's serialized size of {row_bytes / 1000} exceeds max_file_size_kb = {max_file_size_kb}"  # type: ignore[operator]
                    )
                if i == 0:
                    fp, fhb = xstack.enter_context(_start_new_file(file_no))
                    bytes_written = 0
                elif (bytes_written + len(row_bytes)) >= max_file_size:
                    file_no += 1
                    fp, fhb = xstack.enter_context(_start_new_file(file_no))
                    bytes_written = 0

                bytes_written += fhb.write(row_bytes)

        else:
            # count the triples in the graph
            graph_length = len(g)

            if graph_length <= max_triples:
                # the graph is less than max so just NT serialize the whole thing
                g.serialize(
                    destination=Path(output_dir) / f"{file_name_stem}_all.nt",
                    format="nt",
                )
            else:
                # graph_length is > max_lines, make enough files for all graph
                # no_files = math.ceil(graph_length / max_triples)
                file_no = 1 if write_prefixes else 0
                for i, t in enumerate(g.triples((None, None, None))):
                    if i % max_triples == 0:
                        fp, fhb = xstack.enter_context(_start_new_file(file_no))
                        file_no += 1
                    fhb.write(_nt_row(t).encode("utf-8"))
            return



"""
A TriX parser for RDFLib
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, NoReturn, Optional, Tuple
from xml.sax import handler, make_parser
from xml.sax.handler import ErrorHandler

from rdflib.exceptions import ParserError
from rdflib.graph import Graph
from rdflib.namespace import Namespace
from rdflib.parser import InputSource, Parser
from rdflib.store import Store
from rdflib.term import BNode, Identifier, Literal, URIRef

if TYPE_CHECKING:
    # from xml.sax.expatreader import ExpatLocator
    from xml.sax.xmlreader import AttributesImpl, Locator, XMLReader

__all__ = ["create_parser", "TriXHandler", "TriXParser"]


TRIXNS = Namespace("http://www.w3.org/2004/03/trix/trix-1/")
XMLNS = Namespace("http://www.w3.org/XML/1998/namespace")


class TriXHandler(handler.ContentHandler):
    """An Sax Handler for TriX. See http://sw.nokia.com/trix/"""

    lang: Optional[str]
    datatype: Optional[str]

    def __init__(self, store: Store):
        self.store = store
        self.preserve_bnode_ids = False
        self.reset()

    def reset(self) -> None:
        self.bnode: Dict[str, BNode] = {}
        self.graph: Optional[Graph] = None
        self.triple: Optional[List[Identifier]] = None
        self.state = 0
        self.lang = None
        self.datatype = None

    # ContentHandler methods

    def setDocumentLocator(self, locator: Locator):
        self.locator = locator

    def startDocument(self) -> None:
        pass

    def startPrefixMapping(self, prefix: Optional[str], namespace: str) -> None:
        pass

    def endPrefixMapping(self, prefix: Optional[str]) -> None:
        pass

    def startElementNS(
        self, name: Tuple[Optional[str], str], qname, attrs: AttributesImpl
    ) -> None:
        if name[0] != str(TRIXNS):
            self.error(
                "Only elements in the TriX namespace are allowed. %s!=%s"
                % (name[0], TRIXNS)
            )

        if name[1].lower() == "trix":
            if self.state == 0:
                self.state = 1
            else:
                self.error("Unexpected TriX element")

        elif name[1] == "graph":
            if self.state == 1:
                self.state = 2
            else:
                self.error("Unexpected graph element")

        elif name[1] == "uri":
            if self.state == 2:
                # the context uri
                self.state = 3
            elif self.state == 4:
                # part of a triple
                pass
            else:
                self.error("Unexpected uri element")

        elif name[1] == "triple":
            if self.state == 2:
                if self.graph is None:
                    # anonymous graph, create one with random bnode id
                    self.graph = Graph(store=self.store)
                # start of a triple
                self.triple = []
                self.state = 4
            else:
                self.error("Unexpected triple element")

        elif name[1] == "typedLiteral":
            if self.state == 4:
                # part of triple
                self.lang = None
                self.datatype = None

                try:
                    self.lang = attrs.getValue((str(XMLNS), "lang"))  # type: ignore[arg-type, unused-ignore]
                except Exception:
                    # language not required - ignore
                    pass
                try:
                    self.datatype = attrs.getValueByQName("datatype")  # type: ignore[arg-type, unused-ignore]
                except KeyError:
                    self.error("No required attribute 'datatype'")
            else:
                self.error("Unexpected typedLiteral element")

        elif name[1] == "plainLiteral":
            if self.state == 4:
                # part of triple
                self.lang = None
                self.datatype = None
                try:
                    # type error: Argument 1 to "getValue" of "AttributesImpl" has incompatible type "Tuple[str, str]"; expected "str"
                    self.lang = attrs.getValue((str(XMLNS), "lang"))  # type: ignore[arg-type, unused-ignore]
                except Exception:
                    # language not required - ignore
                    pass

            else:
                self.error("Unexpected plainLiteral element")

        elif name[1] == "id":
            if self.state == 2:
                # the context uri
                self.state = 3

            elif self.state == 4:
                # part of triple
                pass
            else:
                self.error("Unexpected id element")

        else:
            self.error("Unknown element %s in TriX namespace" % name[1])

        self.chars = ""

    def endElementNS(self, name: Tuple[Optional[str], str], qname) -> None:
        if TYPE_CHECKING:
            assert self.triple is not None
        if name[0] != str(TRIXNS):
            self.error(
                "Only elements in the TriX namespace are allowed. %s!=%s"
                % (name[0], TRIXNS)
            )

        if name[1] == "uri":
            if self.state == 3:
                self.graph = Graph(
                    store=self.store, identifier=URIRef(self.chars.strip())
                )
                self.state = 2
            elif self.state == 4:
                self.triple += [URIRef(self.chars.strip())]
            else:
                self.error(
                    "Illegal internal self.state - This should never "
                    + "happen if the SAX parser ensures XML syntax correctness"
                )

        elif name[1] == "id":
            if self.state == 3:
                self.graph = Graph(
                    self.store, identifier=self.get_bnode(self.chars.strip())
                )
                self.state = 2
            elif self.state == 4:
                self.triple += [self.get_bnode(self.chars.strip())]
            else:
                self.error(
                    "Illegal internal self.state - This should never "
                    + "happen if the SAX parser ensures XML syntax correctness"
                )

        elif name[1] == "plainLiteral" or name[1] == "typedLiteral":
            if self.state == 4:
                self.triple += [
                    Literal(self.chars, lang=self.lang, datatype=self.datatype)
                ]
            else:
                self.error(
                    "This should never happen if the SAX parser "
                    + "ensures XML syntax correctness"
                )

        elif name[1] == "triple":
            if self.state == 4:
                if len(self.triple) != 3:
                    self.error(
                        "Triple has wrong length, got %d elements: %s"
                        % (len(self.triple), self.triple)
                    )
                # type error: Item "None" of "Optional[Graph]" has no attribute "add"
                # type error: Argument 1 to "add" of "Graph" has incompatible type "List[Identifier]"; expected "Tuple[Node, Node, Node]"
                self.graph.add(self.triple)  # type: ignore[union-attr, arg-type]
                # self.store.store.add(self.triple,context=self.graph)
                # self.store.addN([self.triple+[self.graph]])
                self.state = 2
            else:
                self.error(
                    "This should never happen if the SAX parser "
                    + "ensures XML syntax correctness"
                )

        elif name[1] == "graph":
            self.graph = None
            self.state = 1

        elif name[1].lower() == "trix":
            self.state = 0

        else:
            self.error("Unexpected close element")

    def get_bnode(self, label: str) -> BNode:
        if self.preserve_bnode_ids:
            bn = BNode(label)
        else:
            if label in self.bnode:
                bn = self.bnode[label]
            else:
                bn = BNode(label)
                self.bnode[label] = bn
        return bn

    def characters(self, content: str) -> None:
        self.chars += content

    def ignorableWhitespace(self, content) -> None:
        pass

    def processingInstruction(self, target, data) -> None:
        pass

    def error(self, message: str) -> NoReturn:
        locator = self.locator
        info = "%s:%s:%s: " % (
            locator.getSystemId(),
            locator.getLineNumber(),
            locator.getColumnNumber(),
        )
        raise ParserError(info + message)


def create_parser(store: Store) -> XMLReader:
    parser = make_parser()
    try:
        # Workaround for bug in expatreader.py. Needed when
        # expatreader is trying to guess a prefix.
        # type error: "XMLReader" has no attribute "start_namespace_decl"
        parser.start_namespace_decl("xml", "http://www.w3.org/XML/1998/namespace")  # type: ignore[attr-defined]
    except AttributeError:
        pass  # Not present in Jython (at least)
    parser.setFeature(handler.feature_namespaces, 1)
    trix = TriXHandler(store)
    parser.setContentHandler(trix)
    parser.setErrorHandler(ErrorHandler())
    return parser


class TriXParser(Parser):
    """A parser for TriX. See http://sw.nokia.com/trix/"""

    def __init__(self):
        pass

    def parse(self, source: InputSource, sink: Graph, **args: Any) -> None:
        assert (
            sink.store.context_aware
        ), "TriXParser must be given a context aware store."

        self._parser = create_parser(sink.store)
        content_handler = self._parser.getContentHandler()
        preserve_bnode_ids = args.get("preserve_bnode_ids", None)
        if preserve_bnode_ids is not None:
            # type error: ContentHandler has no attribute "preserve_bnode_ids"
            content_handler.preserve_bnode_ids = preserve_bnode_ids  # type: ignore[attr-defined, unused-ignore]
        # We're only using it once now
        # content_handler.reset()
        # self._parser.reset()
        self._parser.parse(source)



#!/usr/bin/env python
"""
A commandline tool for querying with SPARQL on local files and remote sparql endpoints with custom serialization.

example usage:
```bash
    sq path/to/data.ttl -q "SELECT ?x WHERE {?x a foaf:Person. }"
    rdfpipe test.ttl | sparqlquery - -q "SELECT ?x WHERE {?x a foaf:Person. }" --format json
    sq data.ttl -q "ASK {:john a foaf:Person}" --format xml | grep true
    sq path/to/data.ttl --query-file query.rq
    sq data1.ttl data2.ttl -q "DESCRIBE <http://example.com/john>" --format turtle:+spacious
    sq http://example.com/sparqlendpoint --query-file query.rq
    sq http://example.com/sparqlendpoint --query-file query.rq --username user --password secret
    sq /pyth/to/berkeley.db -q "SELECT ?x WHERE {?x a foaf:Person. }" --remote-storetype BerkeleyDB
```

Tip: You can check the truth value for an ASK query, by regex in stdout for 'true'
or 'false'.
"""
from __future__ import annotations

import argparse
import inspect
import logging
import os
import sys
from inspect import Parameter
from typing import Any, Dict, List, Optional, Tuple, Type
from urllib.parse import urlparse

try:
    # Pyparsing >=3.0.0
    from pyparsing.exceptions import ParseException
except ImportError:
    # Pyparsing 2
    from pyparsing import ParseException

from rdflib.graph import Dataset, Graph
from rdflib.plugin import PluginException
from rdflib.plugin import get as get_plugin
from rdflib.plugin import plugins as get_plugins
from rdflib.query import Result, ResultSerializer
from rdflib.serializer import Serializer
from rdflib.store import Store

from .rdfpipe import _format_and_kws

__all__ = ["sparqlquery"]


class _ArgumentError(Exception):
    pass


class _PrintHelpError(Exception):
    pass


class InvalidQueryError(Exception):
    pass


def sparqlquery(
    endpoints: List[str],
    query: str,
    result_format: Optional[str] = None,
    result_keywords: Dict[str, str] = {},
    auth: Optional[Tuple[str, str]] = None,
    use_stdin: bool = False,
    remote_storetype: Optional[str] = None,
):
    if use_stdin:
        g = Graph().parse(sys.stdin)
    else:
        g = _get_graph(endpoints, auth, remote_storetype)
    try:
        results: Result = g.query(query)
    except ParseException as err:
        raise InvalidQueryError(query) from err

    if result_format is not None:
        ret_bytes = results.serialize(format=result_format, **result_keywords)
    else:
        ret_bytes = results.serialize(**result_keywords)
    if ret_bytes is not None:
        print(ret_bytes.decode())


def _dest_is_local(dest: str):
    if os.path.isabs(dest):
        return True

    q = urlparse(dest)
    # Handle Windows drive letters (single letter followed by colon)
    if len(q.scheme) == 1 and q.scheme.isalpha():
        return True

    return q.scheme in ["", "file"]


def _dest_is_internet_addr(dest: str):
    q = urlparse(dest)
    return q.scheme in ["http", "https"]


def _get_graph(
    endpoints, auth: Optional[Tuple[str, str]], remote_storetype: Optional[str]
) -> Graph:
    graph: Graph
    if remote_storetype is not None:
        storeplugin = get_plugin(remote_storetype, Store)
        if auth:
            store = storeplugin(endpoints[0], auth=auth)  # type: ignore[call-arg]
        else:
            store = storeplugin(endpoints[0])
        graph = Dataset(store)
    else:
        graph = Graph()
        for x in endpoints:
            graph.parse(location=x)
    return graph


def _extract_query_and_format(parser) -> Tuple[Dict[str, Any], Optional[str]]:
    opts: Dict[str, Any] = {}
    tmp_args, rest_args = parser.parse_known_args()
    if tmp_args.query and tmp_args.queryfile is None:
        query = tmp_args.query
    elif tmp_args.queryfile and tmp_args.query is None:
        with open(tmp_args.queryfile, "r") as f:
            query = f.read()
    else:
        query = None

    if query is None:
        construct = False
    elif "DESCRIBE" in query or "CONSTRUCT" in query:
        construct = True
    else:
        construct = False

    if tmp_args.format is not None:
        format_, format_keywords = _format_and_kws(tmp_args.format)
    elif construct:
        format_keywords = {}
        format_ = "turtle"
        construct = True
    else:
        format_keywords = {}
        format_ = "json"
    epilog = _create_epilog_from_format(format_, construct)
    opts = {
        "query": query,
        "result_format": format_,
        "result_keywords": format_keywords,
    }
    return opts, epilog


def parse_args():
    extra_kwargs: Dict[str, Any] = {}
    if sys.version_info > (3, 9):
        extra_kwargs["exit_on_error"] = False
    parser = argparse.ArgumentParser(
        prog="sparqlquery",
        description=__doc__,
        add_help=False,  # add dynamic epilog before help is added
        formatter_class=argparse.RawDescriptionHelpFormatter,
        # else __doc__ wont be printed on error:
        **extra_kwargs,
    )
    parser.add_argument(
        "-q",
        "--query",
        type=str,
        help="Sparql query. Cannot be set together with -qf/--queryfile.",
    )
    parser.add_argument(
        "-qf",
        "--queryfile",
        type=str,
        help="File from where the sparql query is read. "
        "Cannot be set together with -q/--query",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        help="Print sparql result in given format. "
        "Defaults to 'json' on SELECT, to 'xml' on ASK "
        "and to 'turtle' on DESCRIBE and CONSTRUCT. "
        "Keywords as described in epilog can be given "
        "after format like: "
        "FORMAT:(+)KW1,-KW2,KW3=VALUE.",
    )
    opts: Dict[str, Any]
    opts, parser.epilog = _extract_query_and_format(parser)

    parser.add_argument(
        "endpoint",
        nargs="+",
        type=str,
        help="Endpoints for sparql query. "
        "Can be set to multiple files. "
        "Reads from stdin if '-' is given. ",
    )
    parser.add_argument(
        "-w",
        "--warn",
        action="store_true",
        default=False,
        help="Output warnings to stderr " "(by default only critical errors).",
    )
    parser.add_argument(
        "-h",
        "--help",
        # action="store_true",
        # default=False,
        action="help",
        help="show help message and exit. "
        "Also prints information about given format.",
    )
    parser.add_argument(
        "-u", "--username", type=str, help="Username used during authentication."
    )
    parser.add_argument(
        "-p", "--password", type=str, help="Password used during authentication."
    )
    parser.add_argument(
        "-rs",
        "--remote-storetype",
        type=str,
        help="You can specify which storetype should be used. "
        "Can only be set, when using a single endpoint and not stdin. "
        "Will default to 'SparqlStore' when endpoint is an internetaddress.",
    )

    try:  # catch error because exit_on_error=False
        args = parser.parse_args()
    except argparse.ArgumentError as err:
        parser.print_help()
        raise err

    forbidden_format_keywords = [
        x
        for x in opts.get("result_keywords", dict())
        if x in {"self", "stream", "encoding", "format"}
    ]
    if forbidden_format_keywords:
        raise _ArgumentError(
            "'self', 'stream', 'encoding' and 'format' "
            "mustnt be used as keywords for format."
        )

    if opts.get("query") is None:
        parser.print_help()
        raise _ArgumentError("Either -q/--query or -qf/--queryfile must be provided")

    remote_storetype = args.remote_storetype

    if len(args.endpoint) == 1:
        if args.endpoint[0] == "-":
            if remote_storetype is not None:
                raise _ArgumentError(
                    "Cant us remote graphtype, when endpoint is stdin(-)"
                )
            endpoints = []
            opts["use_stdin"] = True
        elif _dest_is_internet_addr(args.endpoint[0]):
            endpoints = args.endpoint
            if remote_storetype is None:
                remote_storetype = "SPARQLStore"
        else:
            endpoints = args.endpoint
    else:
        if remote_storetype is not None:
            raise _ArgumentError(
                "If remote graphtype is set, only a single endpoint is valid."
            )
        endpoints = list(args.endpoint)
        if any(not (_dest_is_local(x)) for x in args.endpoint):
            raise NotImplementedError(
                "If multiple endpoints are given, all must be local files."
            )

    if args.username is not None and args.password is not None:
        if remote_storetype not in ["SPARQLStore"]:
            raise _ArgumentError(
                "Can use password and username only, "
                "when remote-storetype is 'SPARQLStore'."
            )
        opts["auth"] = (args.username, args.password)
    elif args.username is None and args.password is None:
        pass
    else:
        parser.print_help()
        raise _ArgumentError("User only provided one of password and username")

    return endpoints, remote_storetype, args.warn, opts


def _create_epilog_from_format(format_, construct) -> Optional[str]:
    serializer_plugin_type: Type[ResultSerializer | Serializer]
    if construct:
        serializer_plugin_type = Serializer
    else:
        serializer_plugin_type = ResultSerializer
    try:
        plugin = get_plugin(format_, serializer_plugin_type)
    except PluginException:
        available_plugins = [x.name for x in get_plugins(None, ResultSerializer)]
        return (
            f"No plugin registered for sparql result in format '{format_}'. "
            f"available plugins: {available_plugins}"
        )
    serialize_method = plugin.serialize  # type: ignore[attr-defined]
    module = inspect.getmodule(serialize_method)
    if module is None:
        return None
    pydoc_target = ".".join([module.__name__, serialize_method.__qualname__])
    sig = inspect.signature(serialize_method)
    available_keywords = [
        x
        for x, y in sig.parameters.items()
        if y.kind in [Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD]
    ]
    available_keywords.pop(0)  # pop self
    available_keywords.pop(0)  # pop stream
    if serializer_plugin_type == Serializer:
        available_keywords.pop(1)  # pop encoding
    else:
        available_keywords.pop(0)  # pop encoding

    epilog = (
        f"For more customization for format '{format_}' "
        f"use `pydoc {pydoc_target}`. "
    )
    if len(available_keywords) > 0:
        epilog += f"Known keywords are {available_keywords}."
    # there is always **kwargs in .serialize
    epilog += " Further keywords might be valid."
    return epilog


def main():
    try:
        (
            endpoints,
            remote_storetype,
            warn,
            opts,
        ) = parse_args()
    except _PrintHelpError:
        exit()
    except (_ArgumentError, argparse.ArgumentError) as err:
        print(err, file=sys.stderr)
        exit(2)

    if warn:
        loglevel = logging.WARNING
    else:
        loglevel = logging.CRITICAL
    logging.basicConfig(level=loglevel)

    sparqlquery(
        endpoints,
        remote_storetype=remote_storetype,
        **opts,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t
from dataclasses import asdict, dataclass, field
from enum import Enum

from rdflib import Literal, URIRef
from rdflib.util import from_n3


@dataclass(frozen=True)
class TopologyStatus:
    state: str
    primaryTags: dict[str, t.Any]  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        state = t.cast(t.Any, self.state)
        primary_tags = t.cast(t.Any, self.primaryTags)

        if not isinstance(state, str):
            invalid.append(("state", state, type(state)))
        if not isinstance(primary_tags, dict):
            invalid.append(("primaryTags", primary_tags, type(primary_tags)))
        else:
            for key, value in primary_tags.items():
                if not isinstance(key, str):
                    invalid.append((f"primaryTags key '{key}'", key, type(key)))

        if invalid:
            raise ValueError("Invalid TopologyStatus values: ", invalid)


@dataclass(frozen=True)
class RecoveryOperation:
    name: str
    message: str

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        name = t.cast(t.Any, self.name)
        message = t.cast(t.Any, self.message)

        if not isinstance(name, str):
            invalid.append(("name", name, type(name)))
        if not isinstance(message, str):
            invalid.append(("message", message, type(message)))

        if invalid:
            raise ValueError("Invalid RecoveryOperation values: ", invalid)


@dataclass(frozen=True)
class RecoveryStatus:
    state: RecoveryOperation | None = None
    message: str | None = None
    affectedNodes: list[str] = field(default_factory=list)  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        state = t.cast(t.Any, self.state)
        message = t.cast(t.Any, self.message)
        affected_nodes = t.cast(t.Any, self.affectedNodes)

        if state is not None and not isinstance(state, RecoveryOperation):
            invalid.append(("state", state, type(state)))
        if message is not None and not isinstance(message, str):
            invalid.append(("message", message, type(message)))
        if not isinstance(affected_nodes, list):
            invalid.append(("affectedNodes", affected_nodes, type(affected_nodes)))
        else:
            for index, value in enumerate(affected_nodes):
                if not isinstance(value, str):
                    invalid.append((f"affectedNodes[{index}]", value, type(value)))

        if invalid:
            raise ValueError("Invalid RecoveryStatus values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> RecoveryStatus:
        """Create a RecoveryStatus instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.
        Handles empty dict {} by returning a RecoveryStatus with all None/default values.
        The nested 'state' dict (if present) is automatically converted to
        a RecoveryOperation instance.

        Parameters:
            data: A dict containing the recovery status data, typically
                parsed from a JSON response. Can be an empty dict.

        Returns:
            A RecoveryStatus instance with nested dataclass objects.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If nested 'state' cannot be unpacked into RecoveryOperation.
            ValueError: If field validation fails in RecoveryOperation or
                RecoveryStatus (e.g., invalid types).
        """
        # Handle empty dict case
        if not data:
            return cls()

        state = None
        if "state" in data and data["state"] is not None:
            state = RecoveryOperation(**data["state"])

        return cls(
            state=state,
            message=data.get("message"),
            affectedNodes=data.get("affectedNodes", []),
        )


@dataclass(frozen=True)
class NodeStatus:
    address: str
    nodeState: str  # noqa: N815
    term: int
    syncStatus: dict[str, str]  # noqa: N815
    lastLogTerm: int  # noqa: N815
    lastLogIndex: int  # noqa: N815
    endpoint: str
    recoveryStatus: RecoveryStatus | None  # noqa: N815
    topologyStatus: TopologyStatus | None  # noqa: N815
    clusterEnabled: bool | None  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        address = t.cast(t.Any, self.address)
        node_state = t.cast(t.Any, self.nodeState)
        term = t.cast(t.Any, self.term)
        sync_status = t.cast(t.Any, self.syncStatus)
        last_log_term = t.cast(t.Any, self.lastLogTerm)
        last_log_index = t.cast(t.Any, self.lastLogIndex)
        endpoint = t.cast(t.Any, self.endpoint)
        recovery_status = t.cast(t.Any, self.recoveryStatus)
        topology_status = t.cast(t.Any, self.topologyStatus)
        cluster_enabled = t.cast(t.Any, self.clusterEnabled)

        if not isinstance(address, str):
            invalid.append(("address", address, type(address)))
        if not isinstance(node_state, str):
            invalid.append(("nodeState", node_state, type(node_state)))
        if type(term) is not int:
            invalid.append(("term", term, type(term)))
        if not isinstance(sync_status, dict):
            invalid.append(("syncStatus", sync_status, type(sync_status)))
        else:
            for key, value in sync_status.items():
                if not isinstance(key, str):
                    invalid.append((f"syncStatus key '{key}'", key, type(key)))
                if not isinstance(value, str):
                    invalid.append((f"syncStatus['{key}']", value, type(value)))
        if type(last_log_term) is not int:
            invalid.append(("lastLogTerm", last_log_term, type(last_log_term)))
        if type(last_log_index) is not int:
            invalid.append(("lastLogIndex", last_log_index, type(last_log_index)))
        if not isinstance(endpoint, str):
            invalid.append(("endpoint", endpoint, type(endpoint)))
        if recovery_status is not None and not isinstance(
            recovery_status, RecoveryStatus
        ):
            invalid.append(("recoveryStatus", recovery_status, type(recovery_status)))
        if topology_status is not None and not isinstance(
            topology_status, TopologyStatus
        ):
            invalid.append(("topologyStatus", topology_status, type(topology_status)))
        if cluster_enabled is not None and type(cluster_enabled) is not bool:
            invalid.append(("clusterEnabled", cluster_enabled, type(cluster_enabled)))

        if invalid:
            raise ValueError("Invalid NodeStatus values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> NodeStatus:
        """Create a NodeStatus instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.
        The nested 'recoveryStatus' and 'topologyStatus' dicts are automatically
        converted to their respective dataclass instances.

        Parameters:
            data: A dict containing the node status data, typically
                parsed from a JSON response.

        Returns:
            A NodeStatus instance with nested dataclass objects.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If nested dicts cannot be unpacked into their respective
                dataclass instances.
            ValueError: If field validation fails in TopologyStatus, RecoveryOperation,
                RecoveryStatus, or NodeStatus (e.g., invalid types).
        """
        # Handle recoveryStatus - can be empty dict or actual data
        recovery_status = None
        if "recoveryStatus" in data and data["recoveryStatus"]:
            recovery_status = RecoveryStatus.from_dict(data["recoveryStatus"])
        elif "recoveryStatus" in data:
            # Empty dict case
            recovery_status = RecoveryStatus.from_dict({})

        # Handle topologyStatus - optional field
        topology_status = None
        if "topologyStatus" in data and data["topologyStatus"]:
            topology_status = TopologyStatus(**data["topologyStatus"])

        return cls(
            address=data["address"],
            nodeState=data["nodeState"],
            term=data["term"],
            syncStatus=data["syncStatus"],
            lastLogTerm=data["lastLogTerm"],
            lastLogIndex=data["lastLogIndex"],
            endpoint=data["endpoint"],
            recoveryStatus=recovery_status,
            topologyStatus=topology_status,
            clusterEnabled=data.get("clusterEnabled"),
        )


@dataclass(frozen=True)
class ClusterRequest:
    electionMinTimeout: int  # noqa: N815
    electionRangeTimeout: int  # noqa: N815
    heartbeatInterval: int  # noqa: N815
    messageSizeKB: int  # noqa: N815
    verificationTimeout: int  # noqa: N815
    transactionLogMaximumSizeGB: int  # noqa: N815
    batchUpdateInterval: int  # noqa: N815
    nodes: list[str]

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        election_min_timeout = t.cast(t.Any, self.electionMinTimeout)
        election_range_timeout = t.cast(t.Any, self.electionRangeTimeout)
        heartbeat_interval = t.cast(t.Any, self.heartbeatInterval)
        message_size_kb = t.cast(t.Any, self.messageSizeKB)
        verification_timeout = t.cast(t.Any, self.verificationTimeout)
        transaction_log_maximum_size_gb = t.cast(
            t.Any, self.transactionLogMaximumSizeGB
        )
        batch_update_interval = t.cast(t.Any, self.batchUpdateInterval)
        nodes = t.cast(t.Any, self.nodes)

        if type(election_min_timeout) is not int:
            invalid.append(
                ("electionMinTimeout", election_min_timeout, type(election_min_timeout))
            )
        if type(election_range_timeout) is not int:
            invalid.append(
                (
                    "electionRangeTimeout",
                    election_range_timeout,
                    type(election_range_timeout),
                )
            )
        if type(heartbeat_interval) is not int:
            invalid.append(
                ("heartbeatInterval", heartbeat_interval, type(heartbeat_interval))
            )
        if type(message_size_kb) is not int:
            invalid.append(("messageSizeKB", message_size_kb, type(message_size_kb)))
        if type(verification_timeout) is not int:
            invalid.append(
                (
                    "verificationTimeout",
                    verification_timeout,
                    type(verification_timeout),
                )
            )
        if type(transaction_log_maximum_size_gb) is not int:
            invalid.append(
                (
                    "transactionLogMaximumSizeGB",
                    transaction_log_maximum_size_gb,
                    type(transaction_log_maximum_size_gb),
                )
            )
        if type(batch_update_interval) is not int:
            invalid.append(
                (
                    "batchUpdateInterval",
                    batch_update_interval,
                    type(batch_update_interval),
                )
            )

        if not isinstance(nodes, list):
            invalid.append(("nodes", nodes, type(nodes)))
        else:
            for index, value in enumerate(nodes):
                if not isinstance(value, str):
                    invalid.append((f"nodes[{index}]", value, type(value)))

        if invalid:
            raise ValueError("Invalid ClusterRequest values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> ClusterRequest:
        """Create a ClusterRequest instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.

        Parameters:
            data: A dict containing the cluster request data, typically
                parsed from a JSON response.

        Returns:
            A ClusterRequest instance.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If the dict cannot be unpacked into ClusterRequest.
            ValueError: If field validation fails in ClusterRequest
                (e.g., invalid types for integer fields or nodes).
        """
        return cls(
            electionMinTimeout=data["electionMinTimeout"],
            electionRangeTimeout=data["electionRangeTimeout"],
            heartbeatInterval=data["heartbeatInterval"],
            messageSizeKB=data["messageSizeKB"],
            verificationTimeout=data["verificationTimeout"],
            transactionLogMaximumSizeGB=data["transactionLogMaximumSizeGB"],
            batchUpdateInterval=data["batchUpdateInterval"],
            nodes=data["nodes"],
        )


@dataclass(frozen=True)
class SnapshotOptionsBean:
    withRepositoryData: bool  # noqa: N815
    withSystemData: bool  # noqa: N815
    cleanDataDir: bool  # noqa: N815
    repositories: list[str] | None = None

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        with_repository_data = t.cast(t.Any, self.withRepositoryData)
        with_system_data = t.cast(t.Any, self.withSystemData)
        clean_data_dir = t.cast(t.Any, self.cleanDataDir)
        repositories = t.cast(t.Any, self.repositories)

        if type(with_repository_data) is not bool:
            invalid.append(
                ("withRepositoryData", with_repository_data, type(with_repository_data))
            )
        if type(with_system_data) is not bool:
            invalid.append(("withSystemData", with_system_data, type(with_system_data)))
        if type(clean_data_dir) is not bool:
            invalid.append(("cleanDataDir", clean_data_dir, type(clean_data_dir)))

        if repositories is not None:
            if not isinstance(repositories, list):
                invalid.append(("repositories", repositories, type(repositories)))
            else:
                for index, value in enumerate(repositories):
                    if not isinstance(value, str):
                        invalid.append((f"repositories[{index}]", value, type(value)))

        if invalid:
            raise ValueError("Invalid SnapshotOptionsBean values: ", invalid)


@dataclass(frozen=True)
class BackupOperationBean:
    id: str
    username: str
    operation: t.Literal[
        "CREATE_BACKUP_IN_PROGRESS",
        "RESTORE_BACKUP_IN_PROGRESS",
        "CREATE_CLOUD_BACKUP_IN_PROGRESS",
        "RESTORE_CLOUD_BACKUP_IN_PROGRESS",
    ]
    affectedRepositories: list[str]  # noqa: N815
    msSinceCreated: int  # noqa: N815
    snapshotOptions: SnapshotOptionsBean  # noqa: N815
    nodePerformingClusterBackup: str | None = None  # noqa: N815

    def __post_init__(self) -> None:
        _allowed_operations = {
            "CREATE_BACKUP_IN_PROGRESS",
            "RESTORE_BACKUP_IN_PROGRESS",
            "CREATE_CLOUD_BACKUP_IN_PROGRESS",
            "RESTORE_CLOUD_BACKUP_IN_PROGRESS",
        }
        invalid: list[tuple[str, t.Any, type]] = []
        id_ = t.cast(t.Any, self.id)
        username = t.cast(t.Any, self.username)
        operation = t.cast(t.Any, self.operation)
        affected_repositories = t.cast(t.Any, self.affectedRepositories)
        ms_since_created = t.cast(t.Any, self.msSinceCreated)
        snapshot_options = t.cast(t.Any, self.snapshotOptions)
        node_performing_cluster_backup = t.cast(t.Any, self.nodePerformingClusterBackup)

        if not isinstance(id_, str):
            invalid.append(("id", id_, type(id_)))
        if not isinstance(username, str):
            invalid.append(("username", username, type(username)))
        if not isinstance(operation, str) or operation not in _allowed_operations:
            invalid.append(("operation", operation, type(operation)))
        if not isinstance(affected_repositories, list):
            invalid.append(
                (
                    "affectedRepositories",
                    affected_repositories,
                    type(affected_repositories),
                )
            )
        else:
            for index, value in enumerate(affected_repositories):
                if not isinstance(value, str):
                    invalid.append(
                        (f"affectedRepositories[{index}]", value, type(value))
                    )
        if type(ms_since_created) is not int:
            invalid.append(("msSinceCreated", ms_since_created, type(ms_since_created)))
        if not isinstance(snapshot_options, SnapshotOptionsBean):
            invalid.append(
                ("snapshotOptions", snapshot_options, type(snapshot_options))
            )
        if node_performing_cluster_backup is not None and not isinstance(
            node_performing_cluster_backup, str
        ):
            invalid.append(
                (
                    "nodePerformingClusterBackup",
                    node_performing_cluster_backup,
                    type(node_performing_cluster_backup),
                )
            )

        if invalid:
            raise ValueError("Invalid BackupOperationBean values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> BackupOperationBean:
        """Create a BackupOperationBean instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.
        The nested 'snapshotOptions' dict is automatically converted to
        a SnapshotOptionsBean instance.

        Parameters:
            data: A dict containing the backup operation data, typically
                parsed from a JSON response.

        Returns:
            A BackupOperationBean instance with nested dataclass objects.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If nested 'snapshotOptions' cannot be unpacked into
                SnapshotOptionsBean.
            ValueError: If field validation fails in SnapshotOptionsBean or
                BackupOperationBean (e.g., invalid types or operation values).
        """
        snapshot_options = SnapshotOptionsBean(**data["snapshotOptions"])
        return cls(
            id=data["id"],
            username=data["username"],
            operation=data["operation"],
            affectedRepositories=data["affectedRepositories"],
            msSinceCreated=data["msSinceCreated"],
            snapshotOptions=snapshot_options,
            nodePerformingClusterBackup=data.get("nodePerformingClusterBackup"),
        )


@dataclass(frozen=True)
class StructuresStatistics:
    cacheHit: int  # noqa: N815
    cacheMiss: int  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        cache_hit = t.cast(t.Any, self.cacheHit)
        cache_miss = t.cast(t.Any, self.cacheMiss)

        if type(cache_hit) is not int:
            invalid.append(("cacheHit", cache_hit, type(cache_hit)))
        if type(cache_miss) is not int:
            invalid.append(("cacheMiss", cache_miss, type(cache_miss)))

        if invalid:
            raise ValueError("Invalid StructuresStatistics values: ", invalid)


@dataclass(frozen=True)
class RepositoryStatisticsQueries:
    slow: int
    suboptimal: int

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        slow = t.cast(t.Any, self.slow)
        suboptimal = t.cast(t.Any, self.suboptimal)

        if type(slow) is not int:
            invalid.append(("slow", slow, type(slow)))
        if type(suboptimal) is not int:
            invalid.append(("suboptimal", suboptimal, type(suboptimal)))

        if invalid:
            raise ValueError("Invalid RepositoryStatisticsQueries values: ", invalid)


@dataclass(frozen=True)
class RepositoryStatisticsEntityPool:
    epoolReads: int  # noqa: N815
    epoolWrites: int  # noqa: N815
    epoolSize: int  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        epool_reads = t.cast(t.Any, self.epoolReads)
        epool_writes = t.cast(t.Any, self.epoolWrites)
        epool_size = t.cast(t.Any, self.epoolSize)

        if type(epool_reads) is not int:
            invalid.append(("epoolReads", epool_reads, type(epool_reads)))
        if type(epool_writes) is not int:
            invalid.append(("epoolWrites", epool_writes, type(epool_writes)))
        if type(epool_size) is not int:
            invalid.append(("epoolSize", epool_size, type(epool_size)))

        if invalid:
            raise ValueError("Invalid RepositoryStatisticsEntityPool values: ", invalid)


@dataclass(frozen=True)
class RepositoryStatistics:
    queries: RepositoryStatisticsQueries
    entityPool: RepositoryStatisticsEntityPool  # noqa: N815
    activeTransactions: int  # noqa: N815
    openConnections: int  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        queries = t.cast(t.Any, self.queries)
        entity_pool = t.cast(t.Any, self.entityPool)
        active_transactions = t.cast(t.Any, self.activeTransactions)
        open_connections = t.cast(t.Any, self.openConnections)

        if not isinstance(queries, RepositoryStatisticsQueries):
            invalid.append(("queries", queries, type(queries)))
        if not isinstance(entity_pool, RepositoryStatisticsEntityPool):
            invalid.append(("entityPool", entity_pool, type(entity_pool)))
        if type(active_transactions) is not int:
            invalid.append(
                ("activeTransactions", active_transactions, type(active_transactions))
            )
        if type(open_connections) is not int:
            invalid.append(
                ("openConnections", open_connections, type(open_connections))
            )

        if invalid:
            raise ValueError("Invalid RepositoryStatistics values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> RepositoryStatistics:
        """Create a RepositoryStatistics instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.
        The nested 'queries' and 'entityPool' dicts are automatically converted to
        their respective dataclass instances.

        Parameters:
            data: A dict containing the repository statistics data, typically
                parsed from a JSON response.

        Returns:
            A RepositoryStatistics instance with nested dataclass objects.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If nested dicts cannot be unpacked into their respective
                dataclass instances.
            ValueError: If field validation fails in RepositoryStatisticsQueries,
                RepositoryStatisticsEntityPool, or RepositoryStatistics
                (e.g., invalid types).
        """
        queries = RepositoryStatisticsQueries(**data["queries"])
        entity_pool = RepositoryStatisticsEntityPool(**data["entityPool"])
        return cls(
            queries=queries,
            entityPool=entity_pool,
            activeTransactions=data["activeTransactions"],
            openConnections=data["openConnections"],
        )


@dataclass(frozen=True)
class InfrastructureMemoryUsage:
    max: int
    committed: int
    init: int
    used: int

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        max_val = t.cast(t.Any, self.max)
        committed = t.cast(t.Any, self.committed)
        init = t.cast(t.Any, self.init)
        used = t.cast(t.Any, self.used)

        if type(max_val) is not int:
            invalid.append(("max", max_val, type(max_val)))
        if type(committed) is not int:
            invalid.append(("committed", committed, type(committed)))
        if type(init) is not int:
            invalid.append(("init", init, type(init)))
        if type(used) is not int:
            invalid.append(("used", used, type(used)))

        if invalid:
            raise ValueError("Invalid InfrastructureMemoryUsage values: ", invalid)


@dataclass(frozen=True)
class InfrastructureStorageMemory:
    dataDirUsed: int  # noqa: N815
    workDirUsed: int  # noqa: N815
    logsDirUsed: int  # noqa: N815
    dataDirFree: int  # noqa: N815
    workDirFree: int  # noqa: N815
    logsDirFree: int  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        data_dir_used = t.cast(t.Any, self.dataDirUsed)
        work_dir_used = t.cast(t.Any, self.workDirUsed)
        logs_dir_used = t.cast(t.Any, self.logsDirUsed)
        data_dir_free = t.cast(t.Any, self.dataDirFree)
        work_dir_free = t.cast(t.Any, self.workDirFree)
        logs_dir_free = t.cast(t.Any, self.logsDirFree)

        if type(data_dir_used) is not int:
            invalid.append(("dataDirUsed", data_dir_used, type(data_dir_used)))
        if type(work_dir_used) is not int:
            invalid.append(("workDirUsed", work_dir_used, type(work_dir_used)))
        if type(logs_dir_used) is not int:
            invalid.append(("logsDirUsed", logs_dir_used, type(logs_dir_used)))
        if type(data_dir_free) is not int:
            invalid.append(("dataDirFree", data_dir_free, type(data_dir_free)))
        if type(work_dir_free) is not int:
            invalid.append(("workDirFree", work_dir_free, type(work_dir_free)))
        if type(logs_dir_free) is not int:
            invalid.append(("logsDirFree", logs_dir_free, type(logs_dir_free)))

        if invalid:
            raise ValueError("Invalid InfrastructureStorageMemory values: ", invalid)


@dataclass(frozen=True)
class InfrastructureStatistics:
    heapMemoryUsage: InfrastructureMemoryUsage  # noqa: N815
    nonHeapMemoryUsage: InfrastructureMemoryUsage  # noqa: N815
    storageMemory: InfrastructureStorageMemory  # noqa: N815
    threadCount: int  # noqa: N815
    cpuLoad: float  # noqa: N815
    classCount: int  # noqa: N815
    gcCount: int  # noqa: N815
    openFileDescriptors: int  # noqa: N815
    maxFileDescriptors: int  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        heap_memory_usage = t.cast(t.Any, self.heapMemoryUsage)
        non_heap_memory_usage = t.cast(t.Any, self.nonHeapMemoryUsage)
        storage_memory = t.cast(t.Any, self.storageMemory)
        thread_count = t.cast(t.Any, self.threadCount)
        cpu_load = t.cast(t.Any, self.cpuLoad)
        class_count = t.cast(t.Any, self.classCount)
        gc_count = t.cast(t.Any, self.gcCount)
        open_file_descriptors = t.cast(t.Any, self.openFileDescriptors)
        max_file_descriptors = t.cast(t.Any, self.maxFileDescriptors)

        if not isinstance(heap_memory_usage, InfrastructureMemoryUsage):
            invalid.append(
                ("heapMemoryUsage", heap_memory_usage, type(heap_memory_usage))
            )
        if not isinstance(non_heap_memory_usage, InfrastructureMemoryUsage):
            invalid.append(
                (
                    "nonHeapMemoryUsage",
                    non_heap_memory_usage,
                    type(non_heap_memory_usage),
                )
            )
        if not isinstance(storage_memory, InfrastructureStorageMemory):
            invalid.append(("storageMemory", storage_memory, type(storage_memory)))
        if type(thread_count) is not int:
            invalid.append(("threadCount", thread_count, type(thread_count)))
        if type(cpu_load) is not float and type(cpu_load) is not int:
            invalid.append(("cpuLoad", cpu_load, type(cpu_load)))
        if type(class_count) is not int:
            invalid.append(("classCount", class_count, type(class_count)))
        if type(gc_count) is not int:
            invalid.append(("gcCount", gc_count, type(gc_count)))
        if type(open_file_descriptors) is not int:
            invalid.append(
                (
                    "openFileDescriptors",
                    open_file_descriptors,
                    type(open_file_descriptors),
                )
            )
        if type(max_file_descriptors) is not int:
            invalid.append(
                ("maxFileDescriptors", max_file_descriptors, type(max_file_descriptors))
            )

        if invalid:
            raise ValueError("Invalid InfrastructureStatistics values: ", invalid)

    @classmethod
    def from_dict(cls, data: dict) -> InfrastructureStatistics:
        """Create an InfrastructureStatistics instance from a dict.

        This is useful for converting JSON response data into the dataclass structure.
        The nested memory and storage dicts are automatically converted to their
        respective dataclass instances.

        Parameters:
            data: A dict containing the infrastructure statistics data, typically
                parsed from a JSON response.

        Returns:
            An InfrastructureStatistics instance with nested dataclass objects.

        Raises:
            KeyError: If required keys are missing from the input dict.
            TypeError: If nested dicts cannot be unpacked into their respective
                dataclass instances.
            ValueError: If field validation fails in InfrastructureMemoryUsage,
                InfrastructureStorageMemory, or InfrastructureStatistics
                (e.g., invalid types).
        """
        heap_memory_usage = InfrastructureMemoryUsage(**data["heapMemoryUsage"])
        non_heap_memory_usage = InfrastructureMemoryUsage(**data["nonHeapMemoryUsage"])
        storage_memory = InfrastructureStorageMemory(**data["storageMemory"])
        return cls(
            heapMemoryUsage=heap_memory_usage,
            nonHeapMemoryUsage=non_heap_memory_usage,
            storageMemory=storage_memory,
            threadCount=data["threadCount"],
            cpuLoad=float(data["cpuLoad"]),
            classCount=data["classCount"],
            gcCount=data["gcCount"],
            openFileDescriptors=data["openFileDescriptors"],
            maxFileDescriptors=data["maxFileDescriptors"],
        )


@dataclass(frozen=True)
class ParserSettings:
    preserveBNodeIds: bool = False  # noqa: N815
    failOnUnknownDataTypes: bool = False  # noqa: N815
    verifyDataTypeValues: bool = False  # noqa: N815
    normalizeDataTypeValues: bool = False  # noqa: N815
    failOnUnknownLanguageTags: bool = False  # noqa: N815
    verifyLanguageTags: bool = True  # noqa: N815
    normalizeLanguageTags: bool = False  # noqa: N815
    stopOnError: bool = True  # noqa: N815
    contextLink: t.Any | None = None  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        preserve_bnode_ids = t.cast(t.Any, self.preserveBNodeIds)
        fail_on_unknown_data_types = t.cast(t.Any, self.failOnUnknownDataTypes)
        verify_data_type_values = t.cast(t.Any, self.verifyDataTypeValues)
        normalize_data_type_values = t.cast(t.Any, self.normalizeDataTypeValues)
        fail_on_unknown_language_tags = t.cast(t.Any, self.failOnUnknownLanguageTags)
        verify_language_tags = t.cast(t.Any, self.verifyLanguageTags)
        normalize_language_tags = t.cast(t.Any, self.normalizeLanguageTags)
        stop_on_error = t.cast(t.Any, self.stopOnError)

        if type(preserve_bnode_ids) is not bool:
            invalid.append(
                ("preserveBNodeIds", preserve_bnode_ids, type(preserve_bnode_ids))
            )
        if type(fail_on_unknown_data_types) is not bool:
            invalid.append(
                (
                    "failOnUnknownDataTypes",
                    fail_on_unknown_data_types,
                    type(fail_on_unknown_data_types),
                )
            )
        if type(verify_data_type_values) is not bool:
            invalid.append(
                (
                    "verifyDataTypeValues",
                    verify_data_type_values,
                    type(verify_data_type_values),
                )
            )
        if type(normalize_data_type_values) is not bool:
            invalid.append(
                (
                    "normalizeDataTypeValues",
                    normalize_data_type_values,
                    type(normalize_data_type_values),
                )
            )
        if type(fail_on_unknown_language_tags) is not bool:
            invalid.append(
                (
                    "failOnUnknownLanguageTags",
                    fail_on_unknown_language_tags,
                    type(fail_on_unknown_language_tags),
                )
            )
        if type(verify_language_tags) is not bool:
            invalid.append(
                ("verifyLanguageTags", verify_language_tags, type(verify_language_tags))
            )
        if type(normalize_language_tags) is not bool:
            invalid.append(
                (
                    "normalizeLanguageTags",
                    normalize_language_tags,
                    type(normalize_language_tags),
                )
            )
        if type(stop_on_error) is not bool:
            invalid.append(("stopOnError", stop_on_error, type(stop_on_error)))

        # Note: don't check contextLink here since it's of type t.Any.

        if invalid:
            raise ValueError("Invalid ParserSettings values: ", invalid)

    def as_dict(self) -> dict[str, t.Any]:
        return asdict(self)


@dataclass(frozen=True)
class ImportSettings:
    name: str
    status: t.Literal["PENDING", "IMPORTING", "DONE", "ERROR", "NONE", "INTERRUPTING"]
    size: str
    lastModified: int  # noqa: N815
    imported: int
    addedStatements: int  # noqa: N815
    removedStatements: int  # noqa: N815
    numReplacedGraphs: int  # noqa: N815
    message: str = ""
    context: t.Any | None = None
    replaceGraphs: t.List = field(default_factory=list)  # noqa: N815
    baseURI: t.Any | None = None  # noqa: N815
    forceSerial: bool = False  # noqa: N815
    type: str = "file"
    format: t.Any | None = None
    data: t.Any | None = None
    parserSettings: ParserSettings = field(default_factory=ParserSettings)  # noqa: N815

    def __post_init__(self) -> None:
        _allowed_status = {
            "PENDING",
            "IMPORTING",
            "DONE",
            "ERROR",
            "NONE",
            "INTERRUPTING",
        }
        invalid: list[tuple[str, t.Any, type]] = []

        name = t.cast(t.Any, self.name)
        status = t.cast(t.Any, self.status)
        message = t.cast(t.Any, self.message)
        replace_graphs = t.cast(t.Any, self.replaceGraphs)
        force_serial = t.cast(t.Any, self.forceSerial)
        type_ = t.cast(t.Any, self.type)
        parser_settings = t.cast(t.Any, self.parserSettings)
        size = t.cast(t.Any, self.size)
        last_modified = t.cast(t.Any, self.lastModified)
        imported = t.cast(t.Any, self.imported)
        added_statements = t.cast(t.Any, self.addedStatements)
        removed_statements = t.cast(t.Any, self.removedStatements)
        num_replaced_graphs = t.cast(t.Any, self.numReplacedGraphs)

        if not isinstance(name, str):
            invalid.append(("name", name, type(name)))
        if status not in _allowed_status:
            invalid.append(("status", status, type(status)))
        if not isinstance(message, str):
            invalid.append(("message", message, type(message)))
        if not isinstance(replace_graphs, list):
            invalid.append(("replaceGraphs", replace_graphs, type(replace_graphs)))
        if type(force_serial) is not bool:
            invalid.append(("forceSerial", force_serial, type(force_serial)))
        if not isinstance(type_, str):
            invalid.append(("type", type_, type(type_)))
        if not isinstance(parser_settings, ParserSettings):
            invalid.append(("parserSettings", parser_settings, type(parser_settings)))
        if not isinstance(size, str):
            invalid.append(("size", size, type(size)))
        if type(last_modified) is not int:
            invalid.append(("lastModified", last_modified, type(last_modified)))
        if type(imported) is not int:
            invalid.append(("imported", imported, type(imported)))
        if type(added_statements) is not int:
            invalid.append(
                ("addedStatements", added_statements, type(added_statements))
            )
        if type(removed_statements) is not int:
            invalid.append(
                ("removedStatements", removed_statements, type(removed_statements))
            )
        if type(num_replaced_graphs) is not int:
            invalid.append(
                ("numReplacedGraphs", num_replaced_graphs, type(num_replaced_graphs))
            )

        # Note: don't check context, baseURI, format, or data here since they are of type t.Any.

        if invalid:
            raise ValueError("Invalid ImportSettings values: ", invalid)

    def as_dict(self) -> dict[str, t.Any]:
        return asdict(self)


@dataclass(frozen=True)
class ServerImportBody:
    fileNames: list[str]  # noqa: N815
    importSettings: ImportSettings | None = None  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        import_settings = t.cast(t.Any, self.importSettings)
        file_names = t.cast(t.Any, self.fileNames)

        if import_settings is not None and not isinstance(
            import_settings, ImportSettings
        ):
            invalid.append(("importSettings", import_settings, type(import_settings)))
        if not isinstance(file_names, list):
            invalid.append(("fileNames", file_names, type(file_names)))
        else:
            for index, value in enumerate(file_names):
                if not isinstance(value, str):
                    invalid.append((f"fileNames[{index}]", value, type(value)))

        if invalid:
            raise ValueError("Invalid ServerImportBody values: ", invalid)

    def as_dict(self) -> dict[str, t.Any]:
        result = asdict(self)
        if self.importSettings is None:
            result.pop("importSettings", None)
        return result


@dataclass(frozen=True)
class UserUpdate:
    password: str = field(default="")
    appSettings: dict[str, t.Any] = field(default_factory=dict)  # noqa: N815
    gptThreads: list[t.Any] = field(default_factory=list)  # noqa: N815

    def as_dict(self) -> dict[str, t.Any]:
        return asdict(self)


@dataclass(frozen=True)
class UserCreate:
    """Dataclass for creating a new user in GraphDB.

    Unlike `User`, this class does not include `dateCreated` since
    GraphDB automatically assigns this value when the user is created.
    """

    username: str
    password: str
    grantedAuthorities: list[str] = field(default_factory=list)  # noqa: N815
    appSettings: dict[str, t.Any] = field(default_factory=dict)  # noqa: N815
    gptThreads: list[t.Any] = field(default_factory=list)  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        username = t.cast(t.Any, self.username)
        password = t.cast(t.Any, self.password)
        granted_authorities = t.cast(t.Any, self.grantedAuthorities)
        app_settings = t.cast(t.Any, self.appSettings)
        gpt_threads = t.cast(t.Any, self.gptThreads)

        if not isinstance(username, str):
            invalid.append(("username", username, type(username)))
        if not isinstance(password, str):
            invalid.append(("password", password, type(password)))

        if not isinstance(granted_authorities, list):
            invalid.append(
                ("grantedAuthorities", granted_authorities, type(granted_authorities))
            )
        else:
            for index, value in enumerate(granted_authorities):
                if not isinstance(value, str):
                    invalid.append((f"grantedAuthorities[{index}]", value, type(value)))

        if not isinstance(app_settings, dict):
            invalid.append(("appSettings", app_settings, type(app_settings)))
        else:
            for key in app_settings.keys():
                if not isinstance(key, str):
                    invalid.append(("appSettings key", key, type(key)))
                    break

        if not isinstance(gpt_threads, list):
            invalid.append(("gptThreads", gpt_threads, type(gpt_threads)))

        if invalid:
            raise ValueError("Invalid UserCreate values: ", invalid)

    def as_dict(self) -> dict[str, t.Any]:
        return asdict(self)


@dataclass(frozen=True)
class User:
    username: str
    password: str
    dateCreated: int  # noqa: N815
    grantedAuthorities: list[str] = field(default_factory=list)  # noqa: N815
    appSettings: dict[str, t.Any] = field(default_factory=dict)  # noqa: N815
    gptThreads: list[t.Any] = field(default_factory=list)  # noqa: N815

    def __post_init__(self) -> None:
        # Normalize None values from API responses to empty collections
        if self.grantedAuthorities is None:
            object.__setattr__(self, "grantedAuthorities", [])  # type: ignore[unreachable]
        if self.appSettings is None:
            object.__setattr__(self, "appSettings", {})  # type: ignore[unreachable]
        if self.gptThreads is None:
            object.__setattr__(self, "gptThreads", [])  # type: ignore[unreachable]

        invalid: list[tuple[str, t.Any, type]] = []
        username = t.cast(t.Any, self.username)
        password = t.cast(t.Any, self.password)
        date_created = t.cast(t.Any, self.dateCreated)
        granted_authorities = t.cast(t.Any, self.grantedAuthorities)
        app_settings = t.cast(t.Any, self.appSettings)
        gpt_threads = t.cast(t.Any, self.gptThreads)

        if not isinstance(username, str):
            invalid.append(("username", username, type(username)))
        if not isinstance(password, str):
            invalid.append(("password", password, type(password)))
        if not isinstance(date_created, int):
            invalid.append(("dateCreated", date_created, type(date_created)))

        if not isinstance(granted_authorities, list):
            invalid.append(
                ("grantedAuthorities", granted_authorities, type(granted_authorities))
            )
        else:
            for index, value in enumerate(granted_authorities):
                if not isinstance(value, str):
                    invalid.append((f"grantedAuthorities[{index}]", value, type(value)))

        if not isinstance(app_settings, dict):
            invalid.append(("appSettings", app_settings, type(app_settings)))
        else:
            for key in app_settings.keys():
                if not isinstance(key, str):
                    invalid.append(("appSettings key", key, type(key)))
                    break

        if not isinstance(gpt_threads, list):
            invalid.append(("gptThreads", gpt_threads, type(gpt_threads)))

        if invalid:
            raise ValueError("Invalid User values: ", invalid)

    def as_dict(self):
        return asdict(self)


@dataclass(frozen=True)
class AuthenticatedUser:
    """Represents an authenticated user returned from POST /rest/login.

    Attributes:
        username: The username of the authenticated user.
        authorities: List of granted authorities/roles (e.g., ["ROLE_USER", "ROLE_ADMIN"]).
        appSettings: Application settings for the user.
        external: Whether the user is external (e.g., from LDAP/OAuth).
        token: The full Authorization header value (e.g., "GDB <token>").
            Can be passed directly to GraphDBClient's auth parameter.
    """

    username: str
    authorities: list[str] = field(default_factory=list)
    appSettings: dict[str, t.Any] = field(default_factory=dict)  # noqa: N815
    external: bool = False
    token: str = ""

    def __post_init__(self) -> None:
        # Normalize None values from API responses to empty collections
        if self.authorities is None:
            object.__setattr__(self, "authorities", [])  # type: ignore[unreachable]
        if self.appSettings is None:
            object.__setattr__(self, "appSettings", {})  # type: ignore[unreachable]

        invalid: list[tuple[str, t.Any, type]] = []
        username = t.cast(t.Any, self.username)
        authorities = t.cast(t.Any, self.authorities)
        app_settings = t.cast(t.Any, self.appSettings)
        external = t.cast(t.Any, self.external)
        token = t.cast(t.Any, self.token)

        if not isinstance(username, str):
            invalid.append(("username", username, type(username)))
        if not isinstance(authorities, list):
            invalid.append(("authorities", authorities, type(authorities)))
        else:
            for index, value in enumerate(authorities):
                if not isinstance(value, str):
                    invalid.append((f"authorities[{index}]", value, type(value)))
        if not isinstance(app_settings, dict):
            invalid.append(("appSettings", app_settings, type(app_settings)))
        else:
            for key in app_settings.keys():
                if not isinstance(key, str):
                    invalid.append(("appSettings key", key, type(key)))
                    break
        if type(external) is not bool:
            invalid.append(("external", external, type(external)))
        if not isinstance(token, str):
            invalid.append(("token", token, type(token)))

        if invalid:
            raise ValueError("Invalid AuthenticatedUser values: ", invalid)

    @classmethod
    def from_response(cls, data: dict[str, t.Any], token: str) -> AuthenticatedUser:
        """Create an AuthenticatedUser from API response data and token.

        Parameters:
            data: The JSON response body from POST /rest/login.
            token: The GDB token extracted from the Authorization header.

        Returns:
            An AuthenticatedUser instance.

        Raises:
            ValueError: If required fields are missing or invalid.
            TypeError: If data is not a dict.
        """
        if not isinstance(data, dict):
            raise TypeError("Response data must be a dict")
        if "username" not in data:
            raise ValueError("Response data must contain 'username'")

        return cls(
            username=data["username"],
            authorities=data.get("authorities", []),
            appSettings=data.get("appSettings", {}),
            external=data.get("external", False),
            token=token,
        )


@dataclass(frozen=True)
class FreeAccessSettings:
    enabled: bool
    authorities: list[str] = field(default_factory=list)
    appSettings: dict[str, t.Any] = field(default_factory=dict)  # noqa: N815

    def __post_init__(self) -> None:
        invalid: list[tuple[str, t.Any, type]] = []
        enabled = t.cast(t.Any, self.enabled)
        authorities = t.cast(t.Any, self.authorities)
        app_settings = t.cast(t.Any, self.appSettings)
        if type(enabled) is not bool:
            invalid.append(("enabled", enabled, type(enabled)))
        if not isinstance(authorities, list):
            invalid.append(("authorities", authorities, type(authorities)))
        else:
            for index, value in enumerate(authorities):
                if not isinstance(value, str):
                    invalid.append((f"authorities[{index}]", value, type(value)))
        if not isinstance(app_settings, dict):
            invalid.append(("appSettings", app_settings, type(app_settings)))
        else:
            for key in app_settings.keys():
                if not isinstance(key, str):
                    invalid.append(("appSettings key", key, type(key)))
                    break

        if invalid:
            raise ValueError("Invalid FreeAccessSettings values: ", invalid)

    def as_dict(self):
        return asdict(self)


@dataclass(frozen=True)
class RepositorySizeInfo:
    inferred: int
    total: int
    explicit: int

    def __post_init__(self):
        invalid = []
        if type(self.inferred) is not int:
            invalid.append("inferred")
        if type(self.total) is not int:
            invalid.append("total")
        if type(self.explicit) is not int:
            invalid.append("explicit")

        if invalid:
            raise ValueError(
                "Invalid RepositorySizeInfo values: ",
                [(x, self.__dict__[x], type(self.__dict__[x])) for x in invalid],
            )


@dataclass(frozen=True)
class OWLimParameter:
    name: str
    label: str
    value: str


@dataclass(frozen=True)
class RepositoryConfigBean:
    id: str
    title: str
    type: str
    sesameType: str  # noqa: N815
    location: str
    params: dict[str, OWLimParameter] = field(default_factory=dict)


@dataclass(frozen=True)
class RepositoryConfigBeanCreate:
    id: str
    title: str
    type: str
    sesameType: str  # noqa: N815
    location: str
    params: dict[str, OWLimParameter] = field(default_factory=dict)
    missingDefaults: dict[str, OWLimParameter] = field(  # noqa: N815
        default_factory=dict
    )

    def as_dict(self) -> dict:
        """Serialize the dataclass to a Python dict.

        Returns:
            dict: A dictionary representation of the dataclass suitable for use
                with httpx POST requests (e.g., via the `json` parameter).

        Examples:
            >>> config = RepositoryConfigBeanCreate(
            ...     id="test-repo",
            ...     title="Test Repository",
            ...     type="graphdb:FreeSailRepository",
            ...     sesameType="graphdb:FreeSailRepository",
            ...     location="",
            ... )
            >>> config_dict = config.as_dict()
            >>> isinstance(config_dict, dict)
            True
        """
        return asdict(self)


class RepositoryState(str, Enum):
    """Enumeration for repository state values."""

    INACTIVE = "INACTIVE"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    STOPPING = "STOPPING"


@dataclass
class GraphDBRepository:
    id: t.Optional[str] = None
    title: t.Optional[str] = None
    uri: t.Optional[str] = None
    externalUrl: t.Optional[str] = None  # noqa: N815
    local: t.Optional[bool] = None
    type: t.Optional[str] = None
    sesameType: t.Optional[str] = None  # noqa: N815
    location: t.Optional[str] = None
    readable: t.Optional[bool] = None
    writable: t.Optional[bool] = None
    unsupported: t.Optional[bool] = None
    state: t.Optional[RepositoryState] = None

    @classmethod
    def from_dict(cls, data: dict) -> GraphDBRepository:
        """Create a GraphDBRepository instance from a dict.

        Parameters:
            data: A dict containing the repository data, typically
                parsed from a JSON response.

        Returns:
            A GraphDBRepository instance.

        Raises:
            TypeError: If the dict contains keys that do not match
                GraphDBRepository fields.
            ValueError: If the 'state' value is not a valid RepositoryState.
        """
        data_copy = dict(data)
        if "state" in data_copy and data_copy["state"] is not None:
            data_copy["state"] = RepositoryState(data_copy["state"])
        return cls(**data_copy)


@dataclass
class AccessControlEntry:
    scope: t.Literal["statement", "clear_graph", "plugin", "system"]
    policy: t.Literal["allow", "deny", "abstain"]
    role: str

    def as_dict(self) -> dict[str, t.Any]:
        raise NotImplementedError("Use a concrete AccessControlEntry subclass.")

    @classmethod
    def from_dict(
        cls, data: dict
    ) -> (
        SystemAccessControlEntry
        | StatementAccessControlEntry
        | PluginAccessControlEntry
        | ClearGraphAccessControlEntry
    ):
        """Create an AccessControlEntry subclass instance from raw GraphDB data.

        Note: we perform parse validation essentially twice (here and in the
        subclass's __post_init__) to ensure mypy is satisfied with the value's type.

        Parameters:
            data: A dict containing the access control entry data, typically
                parsed from a JSON response.

        Returns:
            An AccessControlEntry subclass instance (SystemAccessControlEntry,
            StatementAccessControlEntry, PluginAccessControlEntry, or
            ClearGraphAccessControlEntry) depending on the 'scope' field.

        Raises:
            TypeError: If data is not a dict.
            ValueError: If 'scope' is not a supported value ('system', 'statement',
                'plugin', 'clear_graph'), or if any field fails validation
                (e.g., invalid policy, operation, role, plugin, subject,
                predicate, object, or graph values).
        """
        if not isinstance(data, dict):
            raise TypeError("ACL entry must be a mapping.")

        scope = data.get("scope")
        if scope == "system":
            policy = _parse_policy(data.get("policy"))
            role = _parse_role(data.get("role"))
            operation = _parse_operation(data.get("operation"))
            return SystemAccessControlEntry(
                scope="system",
                policy=policy,
                role=role,
                operation=operation,
            )
        if scope == "statement":
            policy = _parse_policy(data.get("policy"))
            role = _parse_role(data.get("role"))
            operation = _parse_operation(data.get("operation"))
            subject = _parse_subject(data.get("subject"))
            predicate = _parse_predicate(data.get("predicate"))
            obj = _parse_object(data.get("object"))
            graph = _parse_graph(data.get("context"))
            return StatementAccessControlEntry(
                scope="statement",
                policy=policy,
                role=role,
                operation=operation,
                subject=subject,
                predicate=predicate,
                object=obj,
                graph=graph,
            )
        if scope == "plugin":
            policy = _parse_policy(data.get("policy"))
            role = _parse_role(data.get("role"))
            operation = _parse_operation(data.get("operation"))
            plugin = _parse_plugin(data.get("plugin"))
            return PluginAccessControlEntry(
                scope="plugin",
                policy=policy,
                role=role,
                operation=operation,
                plugin=plugin,
            )
        if scope == "clear_graph":
            policy = _parse_policy(data.get("policy"))
            role = _parse_role(data.get("role"))
            graph = _parse_graph(data.get("context"))
            return ClearGraphAccessControlEntry(
                scope="clear_graph",
                policy=policy,
                role=role,
                graph=graph,
            )

        raise ValueError(f"Unsupported FGAC scope: {scope!r}")


@dataclass
class SystemAccessControlEntry(AccessControlEntry):
    scope: t.Literal["system"]
    policy: t.Literal["allow", "deny", "abstain"]
    role: str
    operation: t.Literal["read", "write", "*"]

    def __post_init__(self) -> None:
        self.policy = _parse_policy(self.policy)
        self.role = _parse_role(self.role)
        self.operation = _parse_operation(self.operation)

    def as_dict(self) -> dict[str, t.Any]:
        return {
            "scope": self.scope,
            "policy": self.policy,
            "role": self.role,
            "operation": self.operation,
        }


@dataclass
class StatementAccessControlEntry(AccessControlEntry):
    scope: t.Literal["statement"]
    policy: t.Literal["allow", "deny", "abstain"]
    role: str
    operation: t.Literal["read", "write", "*"]
    subject: t.Literal["*"] | URIRef
    predicate: t.Literal["*"] | URIRef
    object: t.Literal["*"] | URIRef | Literal
    graph: t.Literal["*", "named", "default"] | URIRef

    def __post_init__(self) -> None:
        self.policy = _parse_policy(self.policy)
        self.role = _parse_role(self.role)
        self.operation = _parse_operation(self.operation)
        self.subject = _parse_subject(self.subject)
        self.predicate = _parse_predicate(self.predicate)
        self.object = _parse_object(self.object)
        self.graph = _parse_graph(self.graph)

    def as_dict(self) -> dict[str, t.Any]:
        return {
            "scope": self.scope,
            "policy": self.policy,
            "role": self.role,
            "operation": self.operation,
            "subject": _format_term(self.subject),
            "predicate": _format_term(self.predicate),
            "object": _format_term(self.object),
            "context": _format_term(self.graph),
        }


@dataclass
class PluginAccessControlEntry(AccessControlEntry):
    scope: t.Literal["plugin"]
    policy: t.Literal["allow", "deny", "abstain"]
    role: str
    operation: t.Literal["read", "write", "*"]
    plugin: str

    def __post_init__(self) -> None:
        self.policy = _parse_policy(self.policy)
        self.role = _parse_role(self.role)
        self.operation = _parse_operation(self.operation)
        self.plugin = _parse_plugin(self.plugin)

    def as_dict(self) -> dict[str, t.Any]:
        return {
            "scope": self.scope,
            "policy": self.policy,
            "role": self.role,
            "operation": self.operation,
            "plugin": self.plugin,
        }


@dataclass
class ClearGraphAccessControlEntry(AccessControlEntry):
    scope: t.Literal["clear_graph"]
    policy: t.Literal["allow", "deny", "abstain"]
    role: str
    graph: t.Literal["*", "named", "default"] | URIRef

    def __post_init__(self) -> None:
        self.policy = _parse_policy(self.policy)
        self.role = _parse_role(self.role)
        self.graph = _parse_graph(self.graph)

    def as_dict(self) -> dict[str, t.Any]:
        return {
            "scope": self.scope,
            "policy": self.policy,
            "role": self.role,
            "context": _format_term(self.graph),
        }


_ALLOWED_POLICIES = {"allow", "deny", "abstain"}
_ALLOWED_OPERATIONS = {"read", "write", "*"}
_ALLOWED_GRAPHS = {"*", "named", "default"}


def _parse_policy(policy: t.Any) -> t.Literal["allow", "deny", "abstain"]:
    if policy not in _ALLOWED_POLICIES:
        raise ValueError(f"Invalid FGAC policy: {policy!r}")
    return t.cast(t.Literal["allow", "deny", "abstain"], policy)


def _parse_operation(operation: t.Any) -> t.Literal["read", "write", "*"]:
    if operation not in _ALLOWED_OPERATIONS:
        raise ValueError(f"Invalid FGAC operation: {operation!r}")
    return t.cast(t.Literal["read", "write", "*"], operation)


def _parse_role(role: t.Any) -> str:
    if not isinstance(role, str):
        raise ValueError(f"Invalid FGAC role: {role!r}")
    return role


def _parse_plugin(plugin: t.Any) -> str:
    if not isinstance(plugin, str):
        raise ValueError(f"Invalid FGAC plugin: {plugin!r}")
    return plugin


def _parse_subject(subject: t.Any) -> t.Literal["*"] | URIRef:
    if subject == "*":
        return "*"
    if isinstance(subject, URIRef):
        return subject
    if isinstance(subject, str):
        parsed = _parse_with_n3(subject)
        if isinstance(parsed, URIRef):
            return parsed
    raise ValueError(f"Invalid FGAC subject: {subject!r}")


def _parse_predicate(predicate: t.Any) -> t.Literal["*"] | URIRef:
    if predicate == "*":
        return "*"
    if isinstance(predicate, URIRef):
        return predicate
    if isinstance(predicate, str):
        parsed = _parse_with_n3(predicate)
        if isinstance(parsed, URIRef):
            return parsed
    raise ValueError(f"Invalid FGAC predicate: {predicate!r}")


def _parse_object(obj: t.Any) -> t.Literal["*"] | URIRef | Literal:
    if obj == "*":
        return "*"
    if isinstance(obj, (URIRef, Literal)):
        return obj
    if isinstance(obj, str):
        parsed = _parse_with_n3(obj)
        if isinstance(parsed, (URIRef, Literal)):
            return parsed
    raise ValueError(f"Invalid FGAC object: {obj!r}")


def _parse_graph(graph: t.Any) -> t.Literal["*", "named", "default"] | URIRef:
    if graph in _ALLOWED_GRAPHS:
        return t.cast(t.Literal["*", "named", "default"], graph)
    if isinstance(graph, URIRef):
        return graph
    if isinstance(graph, str):
        parsed = _parse_with_n3(graph)
        if isinstance(parsed, URIRef):
            return parsed
    raise ValueError(f"Invalid FGAC graph: {graph!r}")


def _parse_with_n3(value: str) -> URIRef | Literal:
    try:
        parsed = from_n3(value)
    except Exception:
        parsed = None
    if isinstance(parsed, (URIRef, Literal)):
        return parsed
    try:
        return URIRef(value)
    except Exception as err:  # pragma: no cover - defensive
        raise ValueError(f"Unable to parse value {value!r} into an RDF term.") from err


def _format_term(value: t.Any) -> str:
    if isinstance(value, (URIRef, Literal)):
        return value.n3()
    return t.cast(str, value)



"""
This is a rdflib plugin for parsing NQuad files into Conjunctive
graphs that can be used and queried. The store that backs the graph
*must* be able to handle contexts.

```python
>>> from rdflib import ConjunctiveGraph, URIRef, Namespace
>>> g = ConjunctiveGraph()
>>> data = open("test/data/nquads.rdflib/example.nquads", "rb")
>>> g.parse(data, format="nquads") # doctest:+ELLIPSIS
<Graph identifier=... (<class 'rdflib.graph.ConjunctiveGraph'>)>
>>> assert len(g.store) == 449
>>> # There should be 16 separate contexts
>>> assert len([x for x in g.store.contexts()]) == 16
>>> # is the name of entity E10009 "Arco Publications"?
>>> #   (in graph http://bibliographica.org/entity/E10009)
>>> # Looking for:
>>> # <http://bibliographica.org/entity/E10009>
>>> #   <http://xmlns.com/foaf/0.1/name>
>>> #   "Arco Publications"
>>> #   <http://bibliographica.org/entity/E10009>
>>> s = URIRef("http://bibliographica.org/entity/E10009")
>>> FOAF = Namespace("http://xmlns.com/foaf/0.1/")
>>> assert(g.value(s, FOAF.name).eq("Arco Publications"))

```
"""

from __future__ import annotations

from codecs import getreader
from typing import Any, MutableMapping, Optional

from rdflib.exceptions import ParserError as ParseError
from rdflib.graph import ConjunctiveGraph, Dataset, Graph
from rdflib.parser import InputSource

# Build up from the NTriples parser:
from rdflib.plugins.parsers.ntriples import W3CNTriplesParser, r_tail, r_wspace
from rdflib.term import BNode

__all__ = ["NQuadsParser"]

_BNodeContextType = MutableMapping[str, BNode]


class NQuadsParser(W3CNTriplesParser):

    # type error: Signature of "parse" incompatible with supertype "W3CNTriplesParser"
    def parse(  # type: ignore[override]
        self,
        inputsource: InputSource,
        sink: Graph,
        bnode_context: Optional[_BNodeContextType] = None,
        skolemize: bool = False,
        **kwargs: Any,
    ):
        """Parse inputsource as an N-Quads file.

        Args:
            inputsource: The source of N-Quads-formatted data.
            sink: The graph where parsed quads will be stored.
            bnode_context: Optional dictionary mapping blank node identifiers to
                [`BNode`][rdflib.term.BNode] instances.
                See `.W3CNTriplesParser.parse` for more details.
            skolemize: Whether to skolemize blank nodes.

        Returns:
            The Dataset containing the parsed quads.

        Raises:
            AssertionError: If the sink store is not context-aware.
            ParseError: If the input is not a file-like object or contains invalid lines.
        """
        assert (
            sink.store.context_aware
        ), "NQuadsParser must be given a context-aware store."
        # Set default_union to True to mimic ConjunctiveGraph behavior
        ds = Dataset(store=sink.store, default_union=True)
        ds_default = ds.default_context  # the DEFAULT_DATASET_GRAPH_ID
        new_default_context = None
        if isinstance(sink, (Dataset, ConjunctiveGraph)):
            new_default_context = sink.default_context
        elif sink.identifier is not None:
            if sink.identifier == ds_default.identifier:
                new_default_context = sink
            else:
                new_default_context = ds.get_context(sink.identifier)

        if new_default_context is not None:
            ds.default_context = new_default_context
            ds.remove_graph(ds_default)  # remove the original unused default graph
        # type error: Incompatible types in assignment (expression has type "ConjunctiveGraph", base class "W3CNTriplesParser" defined the type as "Union[DummySink, NTGraphSink]")
        self.sink: Dataset = ds  # type: ignore[assignment]
        self.skolemize = skolemize

        source = inputsource.getCharacterStream()
        if not source:
            source = inputsource.getByteStream()
            source = getreader("utf-8")(source)

        if not hasattr(source, "read"):
            raise ParseError("Item to parse must be a file-like object.")

        self.file = source
        self.buffer = ""
        while True:
            self.line = __line = self.readline()
            if self.line is None:
                break
            try:
                self.parseline(bnode_context)
            except ParseError as msg:
                raise ParseError("Invalid line (%s):\n%r" % (msg, __line))

        return self.sink

    def parseline(self, bnode_context: Optional[_BNodeContextType] = None) -> None:
        self.eat(r_wspace)
        if (not self.line) or self.line.startswith("#"):
            return  # The line is empty or a comment

        subject = self.subject(bnode_context)
        self.eat(r_wspace)

        predicate = self.predicate()
        self.eat(r_wspace)

        obj = self.object(bnode_context)
        self.eat(r_wspace)

        context = self.uriref() or self.nodeid(bnode_context)
        self.eat(r_tail)

        if self.line:
            raise ParseError("Trailing garbage")
        # Must have a context aware store - add on a normal Graph
        # discards anything where the ctx != graph.identifier
        if context:
            self.sink.get_context(context).add((subject, predicate, obj))
        else:
            self.sink.default_context.add((subject, predicate, obj))



"""A Parser for SPARQL results in XML:

http://www.w3.org/TR/rdf-sparql-XMLres/

Bits and pieces borrowed from:
http://projects.bigasterisk.com/sparqlhttp/

Authors: Drew Perttula, Gunnar Aastrand Grimnes
"""

from __future__ import annotations

import logging
import xml.etree.ElementTree as xml_etree  # noqa: N813
from io import BytesIO
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
    cast,
)
from xml.dom import XML_NAMESPACE
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesNSImpl

from rdflib.query import Result, ResultException, ResultParser, ResultSerializer
from rdflib.term import BNode, Identifier, Literal, URIRef, Variable

try:
    # https://adamj.eu/tech/2021/12/29/python-type-hints-optional-imports/
    import lxml.etree as lxml_etree

    FOUND_LXML = True
except ImportError:
    FOUND_LXML = False

SPARQL_XML_NAMESPACE = "http://www.w3.org/2005/sparql-results#"
RESULTS_NS_ET = "{%s}" % SPARQL_XML_NAMESPACE

log = logging.getLogger(__name__)


class XMLResultParser(ResultParser):
    """A Parser for SPARQL results in XML."""

    # TODO FIXME: content_type should be a keyword only arg.
    def parse(self, source: IO, content_type: Optional[str] = None) -> Result:  # type: ignore[override]
        return XMLResult(source)


class XMLResult(Result):
    def __init__(self, source: IO, content_type: Optional[str] = None):
        parser_encoding: Optional[str] = None
        if hasattr(source, "encoding"):
            if TYPE_CHECKING:
                assert isinstance(source, TextIO)
            parser_encoding = "utf-8"
            source_str = source.read()
            source = BytesIO(source_str.encode(parser_encoding))
        else:
            if TYPE_CHECKING:
                assert isinstance(source, BinaryIO)

        if FOUND_LXML:
            lxml_parser = lxml_etree.XMLParser(huge_tree=True, encoding=parser_encoding)
            tree = cast(
                xml_etree.ElementTree,
                lxml_etree.parse(source, parser=lxml_parser),
            )
        else:
            xml_parser = xml_etree.XMLParser(encoding=parser_encoding)
            tree = xml_etree.parse(source, parser=xml_parser)

        boolean = tree.find(RESULTS_NS_ET + "boolean")
        results = tree.find(RESULTS_NS_ET + "results")

        if boolean is not None:
            type_ = "ASK"
        elif results is not None:
            type_ = "SELECT"
        else:
            raise ResultException("No RDF result-bindings or boolean answer found!")

        Result.__init__(self, type_)

        if type_ == "SELECT":
            self.bindings = []
            for result in results:  # type: ignore[union-attr]
                if result.tag != f"{RESULTS_NS_ET}result":
                    # This is here because with lxml this also gets comments,
                    # not just elements. Also this should not operate on non
                    # "result" elements.
                    continue
                r = {}
                for binding in result:
                    if binding.tag != f"{RESULTS_NS_ET}binding":
                        # This is here because with lxml this also gets
                        # comments, not just elements. Also this should not
                        # operate on non "binding" elements.
                        continue
                    # type error: error: Argument 1 to "Variable" has incompatible type "Union[str, None, Any]"; expected "str"
                    # NOTE on type error: Element.get() can return None, and
                    # this will invariably fail if passed into Variable
                    # constructor as value
                    r[Variable(binding.get("name"))] = parseTerm(binding[0])  # type: ignore[arg-type] # FIXME
                self.bindings.append(r)

            self.vars = [
                # type error: Argument 1 to "Variable" has incompatible type "Optional[str]"; expected "str"
                # NOTE on type error: Element.get() can return None, and this
                # will invariably fail if passed into Variable constructor as
                # value
                Variable(x.get("name"))  # type: ignore[arg-type] # FIXME
                for x in tree.findall(
                    "./%shead/%svariable" % (RESULTS_NS_ET, RESULTS_NS_ET)
                )
            ]

        else:
            self.askAnswer = boolean.text.lower().strip() == "true"  # type: ignore[union-attr]


def parseTerm(element: xml_etree.Element) -> Union[URIRef, Literal, BNode]:
    """rdflib object (Literal, URIRef, BNode) for the given
    elementtree element"""
    tag, text = element.tag, element.text
    if tag == RESULTS_NS_ET + "literal":
        if text is None:
            text = ""
        datatype = None
        lang = None
        if element.get("datatype", None):
            # type error: Argument 1 to "URIRef" has incompatible type "Optional[str]"; expected "str"
            datatype = URIRef(element.get("datatype"))  # type: ignore[arg-type]
        elif element.get("{%s}lang" % XML_NAMESPACE, None):
            lang = element.get("{%s}lang" % XML_NAMESPACE)

        ret = Literal(text, datatype=datatype, lang=lang)

        return ret
    elif tag == RESULTS_NS_ET + "uri":
        # type error: Argument 1 to "URIRef" has incompatible type "Optional[str]"; expected "str"
        return URIRef(text)  # type: ignore[arg-type]
    elif tag == RESULTS_NS_ET + "bnode":
        return BNode(text)
    else:
        raise TypeError("unknown binding type %r" % element)


class XMLResultSerializer(ResultSerializer):
    """Serializes SPARQL results into XML format."""

    def __init__(self, result: Result):
        ResultSerializer.__init__(self, result)

    def serialize(self, stream: IO, encoding: str = "utf-8", **kwargs: Any) -> None:
        writer = SPARQLXMLWriter(stream, encoding)
        if self.result.type == "ASK":
            writer.write_header([])
            # type error: Argument 1 to "write_ask" of "SPARQLXMLWriter" has incompatible type "Optional[bool]"; expected "bool"
            writer.write_ask(self.result.askAnswer)  # type: ignore[arg-type]
        else:
            # type error: Argument 1 to "write_header" of "SPARQLXMLWriter" has incompatible type "Optional[List[Variable]]"; expected "Sequence[Variable]"
            writer.write_header(self.result.vars)  # type: ignore[arg-type]
            writer.write_results_header()
            for b in self.result.bindings:
                writer.write_start_result()
                for key, val in b.items():
                    writer.write_binding(key, val)

                writer.write_end_result()

        writer.close()


# TODO: Rewrite with ElementTree?
class SPARQLXMLWriter:
    """
    Python saxutils-based SPARQL XML Writer
    """

    def __init__(self, output: IO, encoding: str = "utf-8"):
        writer = XMLGenerator(output, encoding)
        writer.startDocument()
        writer.startPrefixMapping("", SPARQL_XML_NAMESPACE)
        writer.startPrefixMapping("xml", XML_NAMESPACE)
        writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "sparql"), "sparql", AttributesNSImpl({}, {})
        )
        self.writer = writer
        self._output = output
        self._encoding = encoding
        self._results = False

    def write_header(self, allvarsL: Sequence[Variable]) -> None:
        self.writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "head"), "head", AttributesNSImpl({}, {})
        )
        for i in range(0, len(allvarsL)):
            attr_vals = {
                (None, "name"): str(allvarsL[i]),
            }
            attr_qnames = {
                (None, "name"): "name",
            }
            self.writer.startElementNS(
                (SPARQL_XML_NAMESPACE, "variable"),
                "variable",
                # type error: Argument 1 to "AttributesNSImpl" has incompatible type "Dict[Tuple[None, str], str]"; expected "Mapping[Tuple[str, str], str]"
                # type error: Argument 2 to "AttributesNSImpl" has incompatible type "Dict[Tuple[None, str], str]"; expected "Mapping[Tuple[str, str], str]"  [arg-type]
                AttributesNSImpl(attr_vals, attr_qnames),  # type: ignore[arg-type]
            )
            self.writer.endElementNS((SPARQL_XML_NAMESPACE, "variable"), "variable")
        self.writer.endElementNS((SPARQL_XML_NAMESPACE, "head"), "head")

    def write_ask(self, val: bool) -> None:
        self.writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "boolean"), "boolean", AttributesNSImpl({}, {})
        )
        self.writer.characters(str(val).lower())
        self.writer.endElementNS((SPARQL_XML_NAMESPACE, "boolean"), "boolean")

    def write_results_header(self) -> None:
        self.writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "results"), "results", AttributesNSImpl({}, {})
        )
        self._results = True

    def write_start_result(self) -> None:
        self.writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "result"), "result", AttributesNSImpl({}, {})
        )
        self._resultStarted = True

    def write_end_result(self) -> None:
        assert self._resultStarted
        self.writer.endElementNS((SPARQL_XML_NAMESPACE, "result"), "result")
        self._resultStarted = False

    def write_binding(self, name: Variable, val: Identifier) -> None:
        assert self._resultStarted

        attr_vals: Dict[Tuple[Optional[str], str], str] = {
            (None, "name"): str(name),
        }
        attr_qnames: Dict[Tuple[Optional[str], str], str] = {
            (None, "name"): "name",
        }
        self.writer.startElementNS(
            (SPARQL_XML_NAMESPACE, "binding"),
            "binding",
            # type error: Argument 1 to "AttributesNSImpl" has incompatible type "Dict[Tuple[None, str], str]"; expected "Mapping[Tuple[str, str], str]"
            # type error: Argument 2 to "AttributesNSImpl" has incompatible type "Dict[Tuple[None, str], str]"; expected "Mapping[Tuple[str, str], str]"
            AttributesNSImpl(attr_vals, attr_qnames),  # type: ignore[arg-type, unused-ignore]
        )

        if isinstance(val, URIRef):
            self.writer.startElementNS(
                (SPARQL_XML_NAMESPACE, "uri"), "uri", AttributesNSImpl({}, {})
            )
            self.writer.characters(val)
            self.writer.endElementNS((SPARQL_XML_NAMESPACE, "uri"), "uri")
        elif isinstance(val, BNode):
            self.writer.startElementNS(
                (SPARQL_XML_NAMESPACE, "bnode"), "bnode", AttributesNSImpl({}, {})
            )
            self.writer.characters(val)
            self.writer.endElementNS((SPARQL_XML_NAMESPACE, "bnode"), "bnode")
        elif isinstance(val, Literal):
            attr_vals = {}
            attr_qnames = {}
            if val.language:
                attr_vals[(XML_NAMESPACE, "lang")] = val.language
                attr_qnames[(XML_NAMESPACE, "lang")] = "xml:lang"
            elif val.datatype:
                attr_vals[(None, "datatype")] = val.datatype
                attr_qnames[(None, "datatype")] = "datatype"

            self.writer.startElementNS(
                (SPARQL_XML_NAMESPACE, "literal"),
                "literal",
                # type error: Argument 1 to "AttributesNSImpl" has incompatible type "Dict[Tuple[Optional[str], str], str]"; expected "Mapping[Tuple[str, str], str]"
                # type error: Argument 2 to "AttributesNSImpl" has incompatible type "Dict[Tuple[Optional[str], str], str]"; expected "Mapping[Tuple[str, str], str]"
                AttributesNSImpl(attr_vals, attr_qnames),  # type: ignore[arg-type, unused-ignore]
            )
            self.writer.characters(val)
            self.writer.endElementNS((SPARQL_XML_NAMESPACE, "literal"), "literal")

        else:
            raise Exception("Unsupported RDF term: %s" % val)

        self.writer.endElementNS((SPARQL_XML_NAMESPACE, "binding"), "binding")

    def close(self) -> None:
        if self._results:
            self.writer.endElementNS((SPARQL_XML_NAMESPACE, "results"), "results")
        self.writer.endElementNS((SPARQL_XML_NAMESPACE, "sparql"), "sparql")
        self.writer.endDocument()
